# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import sys
from collections.abc import Mapping

from gallia.command import BaseCommand
from gallia.commands import HSFZDiscoverer
from gallia.commands.discover.doip import DoIPDiscoverer
from gallia.commands.primitive.generic.pdu import GenericPDUPrimitive
from gallia.commands.primitive.uds.dddi import (
    ClearDynamicallyDefinedDataIdentifierDDDIPrimitive,
    DefineByIdentifierDDDIPrimitive,
    DefineByMemoryAddressDDDIPrimitive,
)
from gallia.commands.primitive.uds.dtc import (
    ClearDTCPrimitive,
    ControlDTCPrimitive,
    ReadDTCPrimitive,
)
from gallia.commands.primitive.uds.ecu_reset import ECUResetPrimitive
from gallia.commands.primitive.uds.iocbi import IOCBIPrimitive
from gallia.commands.primitive.uds.pdu import SendPDUPrimitive
from gallia.commands.primitive.uds.ping import PingPrimitive
from gallia.commands.primitive.uds.rdbi import (
    ReadByIdentifierPrimitive,
)
from gallia.commands.primitive.uds.rmba import RMBAPrimitive
from gallia.commands.primitive.uds.rtcl import RTCLPrimitive
from gallia.commands.primitive.uds.vin import VINPrimitive
from gallia.commands.primitive.uds.wdbi import (
    WriteByIdentifierPrimitive,
)
from gallia.commands.primitive.uds.wmba import WMBAPrimitive
from gallia.commands.scan.uds.identifiers import ScanIdentifiers
from gallia.commands.scan.uds.memory import MemoryFunctionsScanner
from gallia.commands.scan.uds.reset import ResetScanner
from gallia.commands.scan.uds.sa_dump_seeds import SASeedsDumper
from gallia.commands.scan.uds.services import ServicesScanner
from gallia.commands.scan.uds.sessions import SessionsScanner
from gallia.commands.script.rerun import Rerunner
from gallia.commands.script.vecu import DbVirtualECU, RngVirtualECU
from gallia.plugins.plugin import CommandTree, Plugin
from gallia.services.uds import ECU
from gallia.transports import BaseTransport, registry


class UDSPlugin(Plugin):
    @classmethod
    def name(cls) -> str:
        return "Gallia UDS"

    @classmethod
    def description(cls) -> str:
        return "Default Gallia plugin for Unified Diagnostic Services (UDS) functionality"

    @classmethod
    def transports(cls) -> list[type[BaseTransport]]:
        return registry

    @classmethod
    def ecus(cls) -> list[type[ECU]]:
        return [ECU]

    @classmethod
    def commands(cls) -> Mapping[str, CommandTree | type[BaseCommand]]:
        tree = {
            "discover": CommandTree(
                description="discover scanners for hosts and endpoints",
                subtree={"doip": DoIPDiscoverer, "hsfz": HSFZDiscoverer},
            ),
            "primitive": CommandTree(
                description="protocol specific primitives",
                subtree={
                    "uds": CommandTree(
                        description="Universal Diagnostic Services",
                        subtree={
                            "rdbi": ReadByIdentifierPrimitive,
                            "dddi": CommandTree(
                                description="DynamicallyDefineDataIdentifiers",
                                subtree={
                                    "identifier": DefineByIdentifierDDDIPrimitive,
                                    "memory": DefineByMemoryAddressDDDIPrimitive,
                                    "clear": ClearDynamicallyDefinedDataIdentifierDDDIPrimitive,
                                },
                            ),
                            "dtc": CommandTree(
                                description="DiagnosticTroubleCodes",
                                subtree={
                                    "clear": ClearDTCPrimitive,
                                    "control": ControlDTCPrimitive,
                                    "read": ReadDTCPrimitive,
                                },
                            ),
                            "ecu-reset": ECUResetPrimitive,
                            "vin": VINPrimitive,
                            "iocbi": IOCBIPrimitive,
                            "ping": PingPrimitive,
                            "rmba": RMBAPrimitive,
                            "rtcl": RTCLPrimitive,
                            "pdu": SendPDUPrimitive,
                            "wmba": WMBAPrimitive,
                            "wdbi": WriteByIdentifierPrimitive,
                        },
                    ),
                    "generic": CommandTree(
                        description="generic networks primitives",
                        subtree={"pdu": GenericPDUPrimitive},
                    ),
                },
            ),
            "scan": CommandTree(
                description="scanners for network protocol parameters",
                subtree={
                    "uds": CommandTree(
                        description="Universal Diagnostic Services",
                        subtree={
                            "memory": MemoryFunctionsScanner,
                            "reset": ResetScanner,
                            "dump-seeds": SASeedsDumper,
                            "identifiers": ScanIdentifiers,
                            "sessions": SessionsScanner,
                            "services": ServicesScanner,
                        },
                    )
                },
            ),
            "script": CommandTree(
                description="miscellaneous helper scripts",
                subtree={"rerun": Rerunner},
            ),
        }

        if sys.platform.startswith("linux"):
            from gallia.commands.discover.uds.isotp import IsotpDiscoverer
            from gallia.commands.fuzz.uds.pdu import PDUFuzzer

            tree["discover"].subtree.update(
                {
                    "uds": CommandTree(
                        description="Universal Diagnostic Services",
                        subtree={
                            "isotp": IsotpDiscoverer,
                        },
                    ),
                }
            )

            tree.update(
                {
                    "fuzz": CommandTree(
                        description="fuzzing tools",
                        subtree={
                            "uds": CommandTree(
                                description="Universal Diagnostic Services",
                                subtree={"pdu": PDUFuzzer},
                            )
                        },
                    ),
                }
            )

            tree["script"].subtree.update(
                {
                    "vecu": CommandTree(
                        description="spawn a virtual UDS ECU",
                        subtree={"db": DbVirtualECU, "rng": RngVirtualECU},
                    ),
                }
            )

        if sys.platform == "win32":
            from gallia.commands.script.flexray import (
                FRConfigDump,
                FRDump,
            )

            tree["script"].subtree.update(
                {
                    "fr-dump": FRDump,
                    "fr-dump-config": FRConfigDump,
                }
            )

        return tree



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import asyncio
import json
import random
import sys
import traceback
from abc import ABC, abstractmethod
from binascii import hexlify, unhexlify
from copy import copy
from pathlib import Path
from time import time
from typing import Any

import aiosqlite

from gallia.command.config import AutoInt, EnumArg, GalliaBaseModel
from gallia.log import get_logger
from gallia.services.uds.core import service
from gallia.services.uds.core.constants import (
    DataIdentifier,
    EcuResetSubFuncs,
    ReadDTCInformationSubFuncs,
    RoutineControlSubFuncs,
    UDSErrorCodes,
    UDSIsoServices,
)
from gallia.services.uds.core.utils import bytes_repr, int_repr, service_repr, to_bytes
from gallia.services.uds.ecu import ECUState
from gallia.transports import TargetURI

logger = get_logger(__name__)


class UDSServer(ABC):
    class Behavior(GalliaBaseModel):
        default_response_if_service_not_supported: bool = True
        default_response_if_missing_sub_function: bool = True
        default_response_if_sub_function_not_supported: bool = True
        default_response_if_incorrect_format: bool = True
        default_response_if_session_change: bool = True
        default_response_if_session_read: bool = True
        default_response_if_tester_present: bool = True
        default_response_if_none: bool = True
        default_response_if_suppress: bool = True

    def __init__(self, behavior: Behavior | None = None) -> None:
        self.state = ECUState()
        self.state.session = 1
        if behavior is None:
            self.behavior = self.Behavior()
        else:
            self.behavior = behavior

    @property
    @abstractmethod
    def supported_services(
        self,
    ) -> dict[int, dict[UDSIsoServices, list[int] | None]]: ...

    def default_response_if_service_not_supported(
        self, request: service.UDSRequest
    ) -> service.NegativeResponse | None:
        assert self.state.session in self.supported_services, "Virtual ECU in unsupported session"

        if request.service_id not in self.supported_services[self.state.session]:
            if any(request.service_id in s for s in self.supported_services.values()):
                nrc = UDSErrorCodes.serviceNotSupportedInActiveSession
            else:
                nrc = UDSErrorCodes.serviceNotSupported

            return service.NegativeResponse(request.service_id, nrc)

        return None

    def _is_sub_function_service(self, service_id: int) -> bool:
        try:
            service_class = service.UDSService._SERVICES[UDSIsoServices(service_id)]

            return (
                issubclass(service_class, service.SpecializedSubFunctionService)
                or service_class.Request is not None
                and issubclass(service_class.Request, service.SubFunctionRequest)
            )
        except Exception:
            return False

    def _is_sub_function_request(self, request: service.UDSRequest) -> bool:
        return self._is_sub_function_service(request.service_id)

    def default_response_if_missing_sub_function(
        self, request: service.UDSRequest
    ) -> service.NegativeResponse | None:
        try:
            if self._is_sub_function_request(request) and len(request.pdu) < 2:
                return service.NegativeResponse(
                    request.service_id,
                    UDSErrorCodes.incorrectMessageLengthOrInvalidFormat,
                )
        except Exception:
            pass

        return None

    def default_response_if_sub_function_not_supported(
        self, request: service.UDSRequest
    ) -> service.NegativeResponse | None:
        assert self.state.session in self.supported_services, "Virtual ECU in unsupported session"

        # The standards explicitly excludes RoutineControl for this check because the availability of a sub function
        # depends on the routineIdentifier
        if (
            self._is_sub_function_request(request)
            and request.service_id != UDSIsoServices.RoutineControl
        ):
            supported_in_active_session = False
            supported_in_other_session = False

            for session in self.supported_services:
                if UDSIsoServices(request.service_id) not in self.supported_services[session]:
                    continue

                supported_sub_functions = self.supported_services[session][
                    UDSIsoServices(request.service_id)
                ]

                assert supported_sub_functions is not None, (
                    "Sub function services must have a (potentially empty) list of supported sub functions"
                )

                sub_function = request.pdu[1] % 0x80

                if sub_function in supported_sub_functions:
                    if session == self.state.session:
                        supported_in_active_session = True
                        break
                    supported_in_other_session = True

            if not supported_in_active_session:
                if supported_in_other_session:
                    nrc = UDSErrorCodes.subFunctionNotSupportedInActiveSession

                else:
                    nrc = UDSErrorCodes.subFunctionNotSupported

                return service.NegativeResponse(request.service_id, nrc)

        return None

    def default_response_if_incorrect_format(
        self, request: service.UDSRequest
    ) -> service.NegativeResponse | None:
        if isinstance(request, service.RawRequest):
            return service.NegativeResponse(
                request.service_id, UDSErrorCodes.incorrectMessageLengthOrInvalidFormat
            )

        return None

    def default_response_if_session_change(self, request: service.UDSRequest) -> None | (
        service.NegativeResponse | service.DiagnosticSessionControlResponse
    ):
        if isinstance(request, service.DiagnosticSessionControlRequest):
            return service.DiagnosticSessionControlResponse(request.diagnostic_session_type)

        return None

    def default_response_if_session_read(
        self, request: service.UDSRequest
    ) -> service.ReadDataByIdentifierResponse | None:
        if isinstance(request, service.ReadDataByIdentifierRequest):
            if request.data_identifier == DataIdentifier.ActiveDiagnosticSessionDataIdentifier:
                return service.ReadDataByIdentifierResponse(
                    request.data_identifier,
                    to_bytes(self.state.session if self.state.session is not None else 1, 1),
                )

        return None

    def default_response_if_tester_present(
        self, request: service.UDSRequest
    ) -> service.TesterPresentResponse | None:
        if isinstance(request, service.TesterPresentRequest):
            return service.TesterPresentResponse()

        return None

    def default_response_if_none(self, request: service.UDSRequest) -> service.NegativeResponse:
        return service.NegativeResponse(request.service_id, UDSErrorCodes.generalReject)

    def default_response_if_suppress(
        self, request: service.UDSRequest, response: service.UDSResponse
    ) -> service.UDSResponse | None:
        if (
            isinstance(response, service.NegativeResponse)
            or not isinstance(request, service.SubFunctionRequest)
            or not request.suppress_response
        ):
            return response

        return None

    async def update_state(
        self, request: service.UDSRequest, response: service.UDSResponse
    ) -> None:
        if isinstance(response, service.DiagnosticSessionControlResponse):
            self.state.reset()
            self.state.session = response.diagnostic_session_type

        if (
            isinstance(response, service.SecurityAccessResponse)
            and response.security_access_type % 2 == 0
        ):
            self.state.security_access_level = response.security_access_type - 1

        if isinstance(response, service.ECUResetResponse):
            self.state.reset()
            self.state.session = 1

    async def setup(self) -> None:
        pass

    async def teardown(self) -> None:
        pass

    async def respond_without_state_change(
        self, request: service.UDSRequest
    ) -> service.UDSResponse | None:
        response: service.UDSResponse | None

        if (
            self.behavior.default_response_if_service_not_supported
            and (response := self.default_response_if_service_not_supported(request)) is not None
        ):
            return response

        if (
            self.behavior.default_response_if_missing_sub_function
            and (response := self.default_response_if_missing_sub_function(request)) is not None
        ):
            return response

        if (
            self.behavior.default_response_if_sub_function_not_supported
            and (response := self.default_response_if_sub_function_not_supported(request))
            is not None
        ):
            return response

        if (
            self.behavior.default_response_if_incorrect_format
            and (response := self.default_response_if_incorrect_format(request)) is not None
        ):
            return response

        if (
            self.behavior.default_response_if_session_change
            and (response := self.default_response_if_session_change(request)) is not None
        ):
            return response

        if (
            self.behavior.default_response_if_session_read
            and (response := self.default_response_if_session_read(request)) is not None
        ):
            return response

        if (
            self.behavior.default_response_if_tester_present
            and (response := self.default_response_if_tester_present(request)) is not None
        ):
            return response

        if (response := await self.respond_after_default(request)) is not None:
            return response

        if self.behavior.default_response_if_none:
            return self.default_response_if_none(request)

        return None

    async def respond(self, request: service.UDSRequest) -> service.UDSResponse | None:
        response = await self.respond_without_state_change(request)

        if response is not None:
            old_state = copy(self.state)

            await self.update_state(request, response)

            if self.state.__dict__ != old_state.__dict__:
                logger.debug(f"Changed state to {self.state}")

            if self.behavior.default_response_if_suppress:
                return self.default_response_if_suppress(request, response)

        return response

    @abstractmethod
    async def respond_after_default(
        self, request: service.UDSRequest
    ) -> service.UDSResponse | None: ...


class RNG(random.Random):
    def __init__(self, *args: Any):
        super().__init__()

        self.seeds: list[Any] = []
        self.set_seeds(*args)

    def set_seeds(self, *args: Any) -> None:
        self.seeds = list(args)

        if len(self.seeds) == 0:
            self.seed()
        else:
            self.seed("|".join(str(seed) for seed in self.seeds))

    def add_seeds(self, *args: Any) -> None:
        self.set_seeds(*self.seeds, *args)

    def random_bool(self, p_true: float) -> bool:
        return self.random() <= p_true

    def random_payload(self, min_len: int = 0, max_len: int | None = None) -> bytes:
        # Mean length should be a few bytes (here 8)
        byte_length = max(min_len, int(self.expovariate(1 / 8) + 0.5))

        if max_len is not None:
            byte_length = min(max_len, byte_length)

        return bytes(self.randint(0, 255) for _ in range(byte_length))


class RNGEcuState(ECUState):
    def __init__(self) -> None:
        super().__init__()

        self.session = 1
        self.last_sa_response: service.SecurityAccessResponse | None = None

    def reset(self) -> None:
        super().reset()
        self.session = 1
        self.last_sa_response = None


class RandomUDSServer(UDSServer):
    class RandomnessParameters(GalliaBaseModel):
        mandatory_sessions: list[AutoInt] = [1]
        optional_sessions: list[AutoInt] = [2, 3, 4] + list(range(0x40, 0x7F))
        p_session: float = 0.05
        mandatory_services: list[EnumArg[UDSIsoServices]] = [
            UDSIsoServices.DiagnosticSessionControl
        ]
        optional_services: list[EnumArg[UDSIsoServices]] = list(
            set(UDSIsoServices) - set(mandatory_services + [UDSIsoServices.NegativeResponse])
        )
        p_service: float = 0.2
        p_sub_function: float = 0.05
        p_identifier: float = 0.005
        p_correct_payload_format: float = 0.1
        p_dtc_status_mask: float = 0.9

    def __init__(
        self,
        seed: int,
        randomness_parameters: RandomnessParameters | None = None,
        behavior: UDSServer.Behavior | None = None,
    ):
        super().__init__(behavior)

        self.state: RNGEcuState = RNGEcuState()
        self.seed = seed
        self.services: dict[int, dict[UDSIsoServices, list[int] | None]] = {}

        if randomness_parameters is None:
            self.randomness_parameters = self.RandomnessParameters()
        else:
            self.randomness_parameters = randomness_parameters

    async def setup(self) -> None:
        self.randomize()

        logger.notice(f"Initialized random UDS server with seed {self.seed}")
        logger.notice(
            json.dumps(
                {
                    int_repr(session): {
                        f"{int_repr(s.value)} ({service_repr(s)})": str(
                            [int_repr(sf) for sf in sfs]
                        )
                        if sfs is not None
                        else None
                        for s, sfs in services.items()
                    }
                    for session, services in self.services.items()
                },
                indent=4,
                sort_keys=True,
            )
        )

    def randomize(self) -> None:
        rng = RNG(self.seed)

        level = 0
        default_session = 1
        level_sessions = {default_session}
        session_transitions: list[set[int]] = [set() for _ in range(0x7F)]
        session_transitions[default_session] = {default_session}
        combined_sessions = (
            self.randomness_parameters.mandatory_sessions
            + self.randomness_parameters.optional_sessions
        )

        while len(level_sessions) > 0:
            p_transition = (
                self.randomness_parameters.p_session / len(level_sessions) / 2 ** (level + 0.5)
            )
            next_level_sessions = set()
            available_sessions = [
                i
                for i, l in enumerate(session_transitions)  # noqa: E741
                if len(l) > 0
            ]

            for session in level_sessions:
                transitions = [
                    session for session in combined_sessions if rng.random() < p_transition
                ]
                session_transitions[session].update(transitions)
                next_level_sessions.update(transitions)

            for session in next_level_sessions:
                session_transitions[session].add(default_session)

            level_sessions = next_level_sessions - set(available_sessions)
            level += 1

        for session in self.randomness_parameters.mandatory_sessions:
            if len(session_transitions[session]) == 0:
                available_sessions = [
                    i
                    for i, l in enumerate(session_transitions)  # noqa: E741
                    if len(l) > 0
                ]
                session_transitions[rng.choice(available_sessions)].add(session)
                session_transitions[session] = {default_session}

        self.services = {}

        for session, session_specific_transitions in enumerate(session_transitions):
            if len(session_specific_transitions) == 0:
                continue

            self.services[session] = {}

            for supported_service in self.randomness_parameters.mandatory_services + [
                s
                for s in self.randomness_parameters.optional_services
                if rng.random() < self.randomness_parameters.p_service
            ]:
                supported_sub_functions: list[int] | None = None

                if self._is_sub_function_service(supported_service):
                    # For SecurityAccess there are always two consecutive sub functions, the uneven one for RequestSeed,
                    # the even one (+1) for SendKey
                    if supported_service == UDSIsoServices.TesterPresent:
                        supported_sub_functions = [0]
                    elif supported_service == UDSIsoServices.DiagnosticSessionControl:
                        supported_sub_functions = sorted(session_specific_transitions)
                    elif supported_service == UDSIsoServices.SecurityAccess:
                        supported_sub_functions_tmp = [
                            sf
                            for sf in range(1, 0x7E, 2)
                            if rng.random() < self.randomness_parameters.p_sub_function / 2
                        ]
                        supported_sub_functions = []

                        for sf in supported_sub_functions_tmp:
                            supported_sub_functions.append(sf)
                            supported_sub_functions.append(sf + 1)
                    elif supported_service == UDSIsoServices.RoutineControl:
                        supported_sub_functions = [sf.value for sf in RoutineControlSubFuncs]
                    # Currently only this sub function is supported so it doesn't make sense to gamble a lot here
                    elif supported_service == UDSIsoServices.ReadDTCInformation:
                        supported_sub_functions = [ReadDTCInformationSubFuncs.reportDTCByStatusMask]
                    else:
                        supported_sub_functions = [
                            sf
                            for sf in range(1, 0x80, 1)
                            if rng.random() < self.randomness_parameters.p_sub_function
                        ]

                self.services[session][supported_service] = supported_sub_functions

    @property
    def supported_services(
        self,
    ) -> dict[int, dict[UDSIsoServices, list[int] | None]]:
        return self.services

    def stateful_rng(self, *args: Any) -> RNG:
        return RNG(
            str(self.seed) + "|" + str(self.state.session) + "|".join(str(arg) for arg in args)
        )

    async def respond_after_default(
        self, request: service.UDSRequest
    ) -> service.UDSResponse | None:
        # Service specific handling starts here
        # It is assumed, that the service and sub-function, if any, are both supported
        # Furthermore, it is assumed that the request is a valid request for that particular service and sub-function
        if isinstance(request, service.ECUResetRequest):
            return self.ecu_reset(request)
        if isinstance(request, service._SecurityAccessRequest):
            return self.security_access(request)
        if isinstance(request, service.RoutineControlRequest):
            return self.routine_control(request)
        if isinstance(request, service.ReadDataByIdentifierRequest):
            return self.read_data_by_identifier(request)
        if isinstance(request, service.WriteDataByIdentifierRequest):
            return self.write_data_by_identifier(request)
        if isinstance(request, service.InputOutputControlByIdentifierRequest):
            return self.input_output_control_by_identifier(request)
        if isinstance(request, service.ClearDiagnosticInformationRequest):
            return self.clear_diagnostic_information(request)
        if request.service_id == UDSIsoServices.ReadDTCInformation:
            return self.read_dtc_information(request)

        return None

    async def update_state(
        self, request: service.UDSRequest, response: service.UDSResponse
    ) -> None:
        await super().update_state(request, response)

        if not isinstance(response, service.TesterPresentResponse):
            self.state.last_sa_response = (
                response if isinstance(response, service.SecurityAccessResponse) else None
            )

    def ecu_reset(self, request: service.ECUResetRequest) -> service.UDSResponse:
        rng = self.stateful_rng(request.pdu)

        if request.reset_type == EcuResetSubFuncs.enableRapidPowerShutDown:
            return service.ECUResetResponse(request.reset_type, rng.randint(0, 255))

        return service.ECUResetResponse(request.reset_type)

    def security_access(self, request: service._SecurityAccessRequest) -> service.UDSResponse:
        if isinstance(request, service.RequestSeedRequest):
            return service.SecurityAccessResponse(
                request.security_access_type, RNG().random_payload()
            )

        if isinstance(request, service.SendKeyRequest):
            if (
                self.state.last_sa_response is None
                or request.security_access_type
                != self.state.last_sa_response.security_access_type + 1
            ):
                return service.NegativeResponse(
                    request.service_id, UDSErrorCodes.requestSequenceError
                )

            # Let's use the identity as a valid key
            expected_key = self.state.last_sa_response.security_seed
            self.state.last_sa_response = None

            if request.security_key == expected_key:
                return service.SecurityAccessResponse(request.security_access_type)

            return service.NegativeResponse(request.service_id, UDSErrorCodes.invalidKey)

        raise AssertionError

    def routine_control(self, request: service.RoutineControlRequest) -> service.UDSResponse:
        rng = self.stateful_rng(request.service_id, request.routine_identifier)

        if not rng.random_bool(self.randomness_parameters.p_identifier):
            return service.NegativeResponse(request.service_id, UDSErrorCodes.requestOutOfRange)

        rng.add_seeds(request.sub_function)

        if not rng.random_bool(2 / 3):
            return service.NegativeResponse(
                request.service_id, UDSErrorCodes.subFunctionNotSupported
            )

        rng = self.stateful_rng(request.pdu)

        if not rng.random_bool(self.randomness_parameters.p_correct_payload_format):
            return service.NegativeResponse(
                request.service_id, UDSErrorCodes.incorrectMessageLengthOrInvalidFormat
            )

        return request.RESPONSE_TYPE(request.routine_identifier, rng.random_payload())  # type: ignore

    def read_data_by_identifier(
        self, request: service.ReadDataByIdentifierRequest
    ) -> service.UDSResponse:
        rng = self.stateful_rng(request.pdu)

        if not rng.random_bool(self.randomness_parameters.p_identifier):
            return service.NegativeResponse(request.service_id, UDSErrorCodes.requestOutOfRange)

        return service.ReadDataByIdentifierResponse(
            request.data_identifier, rng.random_payload(min_len=1)
        )

    def write_data_by_identifier(
        self, request: service.WriteDataByIdentifierRequest
    ) -> service.UDSResponse:
        rng = self.stateful_rng(request.service_id, request.data_identifier)

        if not rng.random_bool(self.randomness_parameters.p_identifier):
            return service.NegativeResponse(request.service_id, UDSErrorCodes.requestOutOfRange)

        rng = self.stateful_rng(request.pdu)

        if not rng.random_bool(self.randomness_parameters.p_correct_payload_format):
            return service.NegativeResponse(
                request.service_id, UDSErrorCodes.incorrectMessageLengthOrInvalidFormat
            )

        return service.WriteDataByIdentifierResponse(request.data_identifier)

    def input_output_control_by_identifier(
        self, request: service.InputOutputControlByIdentifierRequest
    ) -> service.UDSResponse:
        rng = self.stateful_rng(request.service_id, request.data_identifier)

        if not rng.random_bool(self.randomness_parameters.p_identifier):
            return service.NegativeResponse(request.service_id, UDSErrorCodes.requestOutOfRange)

        rng = self.stateful_rng(request.pdu)

        if not rng.random_bool(self.randomness_parameters.p_correct_payload_format):
            return service.NegativeResponse(
                request.service_id, UDSErrorCodes.incorrectMessageLengthOrInvalidFormat
            )

        return request.RESPONSE_TYPE(request.data_identifier, rng.random_payload(min_len=1))  # type: ignore

    def clear_diagnostic_information(
        self, request: service.ClearDiagnosticInformationRequest
    ) -> service.UDSResponse:
        rng = self.stateful_rng(request.service_id, request.group_of_dtc)

        if not rng.random_bool(self.randomness_parameters.p_dtc_status_mask):
            return service.NegativeResponse(request.service_id, UDSErrorCodes.requestOutOfRange)

        return service.ClearDiagnosticInformationResponse()

    def read_dtc_information(self, request: service.UDSRequest) -> service.UDSResponse:
        assert request.service_id == UDSIsoServices.ReadDTCInformation

        # Currently only this sub function is supported
        if isinstance(request, service.ReportDTCByStatusMaskRequest):
            # The supported dtc status bits should be equal for all requests
            # and only differ among different ECUs and states
            dtc_status_availability_mask = self.stateful_rng().randint(0, 255)

            rng = self.stateful_rng(request.service_id, request.dtc_status_mask)
            dtc_and_status_record = {}

            # On average around 50 dtcs should be fine
            for _ in range(int(rng.expovariate(1 / 50) + 0.5)):
                dtc_and_status_record[rng.randint(0, 256**3 - 1)] = (
                    rng.randint(0, 255) & dtc_status_availability_mask
                )

            return service.ReportDTCByStatusMaskResponse(
                dtc_status_availability_mask, dtc_and_status_record
            )

        return service.NegativeResponse(request.service_id, UDSErrorCodes.subFunctionNotSupported)


class DBUDSServer(UDSServer):
    class Behavior(UDSServer.Behavior):
        default_response_if_service_not_supported: bool = False
        default_response_if_missing_sub_function: bool = False
        default_response_if_sub_function_not_supported: bool = False
        default_response_if_incorrect_format: bool = False
        default_response_if_session_change: bool = False
        default_response_if_session_read: bool = False
        default_response_if_tester_present: bool = False
        default_response_if_none: bool = False
        default_response_if_suppress: bool = False

    def __init__(
        self,
        db_path: Path,
        ecu: str | None,
        properties: dict[str, Any] | None,
        behavior: Behavior | None = None,
    ):
        super().__init__(behavior)

        self.db_path = db_path
        self.ecu = ecu
        self.properties = properties
        self.connection: aiosqlite.Connection | None = None
        self.last_response = -1

    async def setup(self) -> None:
        self.connection = await aiosqlite.connect(self.db_path)

    async def teardown(self) -> None:
        if self.connection is not None:
            await self.connection.close()

    @property
    def supported_services(
        self,
    ) -> dict[int, dict[UDSIsoServices, list[int] | None]]:
        return {}

    async def respond_after_default(
        self, request: service.UDSRequest
    ) -> service.UDSResponse | None:
        assert self.connection is not None

        query = "SELECT r.id, r.response_pdu FROM scan_result r WHERE "
        parameters: list[Any] = []

        if self.properties is not None and self.ecu is None:
            query = (
                "SELECT r.id, r.response_pdu FROM scan_result r, scan_run s WHERE r.run = s.id AND "
            )

        if self.ecu is not None:
            query = (
                "SELECT r.id, r.response_pdu "
                "FROM scan_run s, scan_result r, address a, ecu e "
                "WHERE r.run = s.id AND s.address = a.id AND a.ecu = e.id "
                "AND e.name = ? AND "
            )

            parameters = [self.ecu]

        for key, value in self.state.__dict__.items():
            if value is None:
                query += f"json_extract(r.state, '$.{key}') IS NULL AND "
            else:
                query += f"json_extract(r.state, '$.{key}') = ? AND "
                parameters.append(value if isinstance(value, int | float) else json.dumps(value))

        if self.properties is not None:
            for key, value in self.properties.items():
                if value is None:
                    query += f"json_extract(s.properties_pre, '$.{key}') IS NULL AND "
                else:
                    query += f"json_extract(s.properties_pre, '$.{key}') = ? AND "
                    parameters.append(
                        value if isinstance(value, int | float) else json.dumps(value)
                    )

        query += "r.request_pdu = ? "

        final_query = f"{query} AND r.id > ? ORDER BY r.id LIMIT 1 "

        parameters += [bytes_repr(request.pdu, False, None), self.last_response]
        cursor: aiosqlite.Cursor = await self.connection.execute(final_query, parameters)
        result = await cursor.fetchone()

        if result is None:
            final_query = f"{query} AND r.id <= ? ORDER BY r.id LIMIT 1 "
            cursor = await self.connection.execute(final_query, parameters)
            result = await cursor.fetchone()

        if result is not None:
            self.last_response = result[0]
            response_pdu: bytes | None = result[1]

            if response_pdu is not None:
                response = service.UDSResponse.parse_dynamic(unhexlify(response_pdu))
                return response

            logger.info("Reset ECU due to missing response")
            self.state.reset()
            self.state.session = 1

        return None


class UDSServerTransport:
    def __init__(self, server: UDSServer, target: TargetURI):
        self.server = server
        self.target = target
        self.last_time_active = time()

    async def run(self) -> None: ...

    async def handle_request(self, request_pdu: bytes) -> tuple[bytes | None, float]:
        start = time()

        if start - self.last_time_active > 10:
            logger.info("Server state reset due to inactivity")
            self.server.state.reset()

        request = service.UDSRequest.parse_dynamic(request_pdu)
        logger.debug(f"---> {request}")
        response = await self.server.respond(request)
        end = time()
        self.last_time_active = end

        if response is not None:
            logger.debug(f"  <--- {response} after {(end - start) * 1000:.2f} ms")
            return response.pdu, end - start

        logger.debug(f"  x--- NO RESPONSE after {(end - start) * 1000:.2f} ms")
        return None, end - start


class TCPUDSServerTransport(UDSServerTransport):
    async def handle_client(
        self,
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
    ) -> None:
        logger.info("New connection")
        response_times = []

        while True:
            try:
                line = await reader.readline()

                if not line:
                    break

                tcp_request = line.decode("ascii").strip()
                uds_request_raw = unhexlify(tcp_request)
                uds_response_raw, response_time = await self.handle_request(uds_request_raw)
                response_times.append(response_time)

                if uds_response_raw is not None:
                    writer.write(hexlify(uds_response_raw) + b"\n")
                    await writer.drain()
            except Exception as e:
                logger.error(f"Unexpected exception when handling client communication: {e!r}")
                traceback.print_exc()
                break

        logger.info("Connection closed")
        if len(response_times) > 0:
            logger.info(
                f"Average response time: {sum(response_times) / len(response_times) * 1000:.2f}ms"
            )

    async def run(self) -> None:
        server = await asyncio.start_server(
            self.handle_client, self.target.hostname, self.target.port
        )

        async with server:
            await server.serve_forever()


if sys.platform.startswith("linux"):
    from gallia.transports import ISOTPTransport

    class ISOTPUDSServerTransport(UDSServerTransport):
        async def run(self) -> None:
            transport = await ISOTPTransport.connect(self.target)

            while True:
                try:
                    uds_request_raw = await transport.read()
                    uds_response_raw, _ = await self.handle_request(uds_request_raw)

                    if uds_response_raw is not None:
                        await transport.write(uds_response_raw)
                except Exception as e:
                    logger.error(f"Unexpected exception when handling client communication: {e!r}")
                    traceback.print_exc()
                    break


if sys.platform.startswith("linux") or sys.platform == "darwin":

    class UnixUDSServerTransport(TCPUDSServerTransport):
        async def run(self) -> None:
            server = await asyncio.start_unix_server(self.handle_client, self.target.path)

            async with server:
                await server.serve_forever()



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

from typing import Any, Unpack

from pydantic.fields import FieldInfo, _FromFieldInfoInputs
from pydantic_core import PydanticUndefined


class ArgFieldInfo(FieldInfo):
    def __init__(
        self,
        default: Any,
        positional: bool,
        short: str | None,
        metavar: str | None,
        cli_group: str | None,
        const: Any,
        hidden: bool,
        **kwargs: Unpack[_FromFieldInfoInputs],
    ):
        """Creates a new ArgFieldInfo.

        This is a special variant of pydantic's FieldInfo, which adds several arguments,
        mainly related to CLI arguments.
        For general usage and details on the generic parameters see https://docs.pydantic.dev/latest/concepts/fields.
        Just as with pydantic's FieldInfo, this should usually not be called directly.
        Instead use the Field() function of this module.

        :param default: The default value, if non is given explicitly.
        :param positional: Specifies, if the argument is shown as positional, as opposed to optional (default), on the CLI.
        :param short: An optional alternative name for the CLI, which is auto-prefixed with "-".
        :param metavar: The type hint which is shown on the CLI for an argument. If none is specified, it is automatically inferred from its type.
        :param cli_group: The group in the CLI under which the argument is listed.
        :param const: Specifies, a default value, if the argument is set with no explicit value.
        :param hidden: Specifies, that the argument is part of neither the CLI nor the config file.
        :param kwargs: Generic pydantic Field() arguments (see https://docs.pydantic.dev/latest/api/fields/#pydantic.fields.FieldInfo).
        """
        super().__init__(default=default, **kwargs)

        self.positional = positional
        self.short = short
        self.metavar = metavar
        self.group = cli_group
        self.const = const
        self.hidden = hidden


def Field(
    default: Any = PydanticUndefined,
    positional: bool = False,
    short: str | None = None,
    metavar: str | None = None,
    group: str | None = None,
    const: Any = PydanticUndefined,
    hidden: bool = False,
    **kwargs: Unpack[_FromFieldInfoInputs],
) -> Any:
    """Creates a new ArgFieldInfo.

    This is a special variant of pydantic's Field() function, which adds several arguments,
    mainly related to CLI arguments.
    For general usage and details on the generic parameters see https://docs.pydantic.dev/latest/concepts/fields.

    :param default: The default value, if non is given explicitly.
    :param positional: Specifies, if the argument is shown as positional, as opposed to optional (default), on the CLI.
    :param short: An optional alternative name for the CLI, which is auto-prefixed with "-".
    :param metavar: The type hint which is shown on the CLI for an argument. If none is specified, it is automatically inferred from its type.
    :param cli_group: The group in the CLI under which the argument is listed.
    :param const: Specifies, a default value, if the argument is set with no explicit value.
    :param hidden: Specifies, that the argument is part of neither the CLI nor the config file.
    :param kwargs: Generic pydantic Field() arguments (see https://docs.pydantic.dev/latest/api/fields/#pydantic.fields.Field).
    :return: A ConfigArgFieldInfo.
    """
    return ArgFieldInfo(default, positional, short, metavar, group, const, hidden, **kwargs)



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import argparse
import asyncio
import json
import os
import sys
from collections.abc import Callable, Mapping, MutableMapping, Sequence
from importlib.metadata import version as meta_version
from pprint import pprint
from textwrap import wrap
from types import UnionType
from typing import Any, Never

import argcomplete
from pydantic import Field, create_model
from pydantic_core import PydanticUndefined

from gallia import exitcodes
from gallia.command import BaseCommand
from gallia.command.base import BaseCommandConfig
from gallia.command.config import GalliaBaseModel
from gallia.config import Config, load_config_file
from gallia.log import setup_logging
from gallia.plugins.plugin import CommandTree, load_commands, load_plugins
from gallia.pydantic_argparse import ArgumentParser
from gallia.pydantic_argparse import BaseCommand as PydanticBaseCommand
from gallia.utils import get_log_level

defaults = dict[type, dict[str, Any]]
_CLASS_ATTR = "_dynamic_gallia_command_class_reference"


def _create_parser_from_command(
    command: type[BaseCommand], config: Config, extra_defaults: defaults, model_counter: int = 0
) -> tuple[type[PydanticBaseCommand], defaults, int]:
    config_attributes = command.CONFIG_TYPE.attributes_from_config(config)
    env_attributes = command.CONFIG_TYPE.attributes_from_env()
    config_attributes.update(env_attributes)

    # it is necessary to create a submodel, because several commands can use the same config
    # (e.g. of their base class if no additional arguments are required)
    model_type = create_model(
        f"_dynamic_{command.CONFIG_TYPE}_{model_counter}", __base__=command.CONFIG_TYPE
    )
    extra_defaults[model_type] = config_attributes

    setattr(model_type, _CLASS_ATTR, command)

    return model_type, extra_defaults, model_counter


def _create_parser_from_tree(
    command_tree: CommandTree, config: Config, extra_defaults: defaults, model_counter: int = 0
) -> tuple[type[PydanticBaseCommand], defaults, int]:
    model_name = f"_dynamic_gallia_hierarchy_model_{model_counter}"
    args: MutableMapping[str, tuple[type | UnionType, Any]] = {}

    for key, value in command_tree.subtree.items():
        model_counter += 1

        if isinstance(value, CommandTree):
            model_type, extra_defaults, model_counter = _create_parser_from_tree(
                value, config, extra_defaults, model_counter
            )
            description = value.description
        else:
            model_type, extra_defaults, model_counter = _create_parser_from_command(
                value, config, extra_defaults, model_counter
            )
            description = value.SHORT_HELP

        args[key] = (model_type | None, Field(None, description=description))

    return (
        create_model(model_name, __base__=PydanticBaseCommand, **args),  # type: ignore[call-overload]
        extra_defaults,
        model_counter,
    )


# TODO: Move this function into some CLI library package.
def create_parser(
    commands: type[BaseCommand] | MutableMapping[str, CommandTree | type[BaseCommand]],
) -> ArgumentParser[PydanticBaseCommand]:
    """
    Creates an argument parser out of the given command hierarchy.

    For accessing the command after parsing, see get_command().
    See parse_and_run() for an easy-to-use one call alternative.

    :param commands: A hierarchy of commands.
    :return The argument parser for the given commands.
    """

    config, _ = load_config_file()

    if isinstance(commands, Mapping):
        command_tree = CommandTree("", subtree=commands)
        model, extra_defaults, _ = _create_parser_from_tree(command_tree, config, {})
    else:
        model, extra_defaults, _ = _create_parser_from_command(commands, config, {})

    return ArgumentParser(model=model, extra_defaults=extra_defaults)


def get_command(config: BaseCommandConfig) -> BaseCommand:
    """
    Retrieve the command out of the config returned by an argument parser as created by create_parser().

    :param config:
    :return: The command initiated with the given config.
    """
    cmd: type[BaseCommand] = getattr(config, _CLASS_ATTR)

    return cmd(config)


# TODO: Move this function into some CLI library package.
def parse_and_run(
    commands: type[BaseCommand] | MutableMapping[str, CommandTree | type[BaseCommand]],
    auto_complete: bool = True,
    setup_log: bool = True,
    top_level_options: Mapping[str, tuple[Callable[[], None], str]] | None = None,
    show_help_on_zero_args: bool = True,
) -> Never:
    """
    Creates an argument parser out of the given command hierarchy and runs the command with its argument.

    This function never returns.
    A set of commands is simply generated by a dict with command name as key and a Command object as value.
    For a full hierarchy of commands CommandTrees can be used.

    :param commands: A hierarchy of commands.
    :param auto_complete: Turns auto-complete functionality on.
    :param setup_log: Setup logging according to the parameters in the parsed config.
    :param top_level_options: Optional top-level actions, such as "--version", given by a mapping of arguments and
                              functions. The program redirects control to the given function, once the program is
                              called with the corresponding argument and terminates after it returns.
    :param show_help_on_zero_args: Show the help message instead of an error in case no arguments are submitted at all.
    """
    parser = create_parser(commands)

    if top_level_options is not None:
        for name, (func, help_) in top_level_options.items():

            class Action(argparse.Action):
                cmd = staticmethod(func)

                def __call__(
                    self,
                    parser: argparse.ArgumentParser,
                    namespace: argparse.Namespace,
                    values: str | Sequence[Any] | None,
                    option_string: str | None = None,
                ) -> None:
                    self.cmd()
                    sys.exit(exitcodes.OK)

            parser.add_argument(
                name,
                nargs=0,
                action=Action,
                help=help_,
            )

    if auto_complete:
        argcomplete.autocomplete(parser)

    if show_help_on_zero_args and len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(exitcodes.USAGE)

    _, config = parser.parse_typed_args()

    assert isinstance(config, BaseCommandConfig)

    if setup_log:
        setup_logging(
            level=get_log_level(config.verbose),
            no_volatile_info=not config.volatile_info,
            logger_name="",  # Take over the root logger
        )

    sys.exit(asyncio.run(get_command(config).entry_point()))


def version() -> None:
    """
    Prints the currently installed version of gallia.
    """
    print(f"gallia {meta_version('gallia')}")


def _walk_commands(
    commands: Mapping[str, CommandTree | type[BaseCommand]], level: int = 0
) -> tuple[str, int]:
    command_str = ""
    command_ctr = 0

    for name, command in commands.items():
        command_str += f"{'  ' * (level + 2)}{name}"

        if isinstance(command, CommandTree):
            command_str += "\n"

            sub_command_str, sub_command_ctr = _walk_commands(command.subtree, level + 1)
            command_ctr += sub_command_ctr
            command_str += sub_command_str
        else:
            command_ctr += 1
            command_str += f": {command.__module__}.{command.__name__}\n"

    return command_str, command_ctr


def show_plugins() -> None:
    """
    Prints the currently installed plugins.
    """
    plugins = load_plugins()

    print(f"There are currently {len(plugins)} plugins installed:")

    for plugin in plugins:
        print()
        print(plugin.name())

        ecus = plugin.ecus()
        print(f"  ECUs ({len(ecus)}):")

        for ecu in ecus:
            print(f"    {ecu.OEM}: {ecu.__module__}.{ecu.__name__}")

        transports = plugin.transports()
        print(f"\n  Transports ({len(transports)}):")

        for transport in transports:
            print(f"    {transport.SCHEME}: {transport.__module__}.{transport.__name__}")

        commands = plugin.commands()
        command_str, command_ctr = _walk_commands(commands)

        print(f"\n  Commands ({command_ctr}):")
        print(command_str)


def show_config() -> None:
    """
    Prints the currently loaded config.
    """
    if (p := os.getenv("GALLIA_CONFIG")) is not None:
        print(f"path to config set by env variable: {p}", file=sys.stderr)

    config, config_path = load_config_file()

    if config_path is not None:
        print(f"loaded config: {config_path}", file=sys.stderr)
        pprint(config)
    else:
        print("no config available", file=sys.stderr)
        sys.exit(exitcodes.CONFIG)


def template() -> None:
    """
    Prints a template config with default according to the programmatic defaults.
    """
    groups: dict[str, dict[str, tuple[str, Any]]] = {}

    for key, value in GalliaBaseModel.registry().items():
        tmp = key.split(".")
        group = ".".join(tmp[:-1])
        attribute = tmp[-1]

        if group not in groups:
            groups[group] = {}

        groups[group][attribute] = value

    output = ""

    for group in sorted(groups):
        if group != "":
            output += f"[{group}]\n"

        for attribute in sorted(groups[group]):
            description, default_value = groups[group][attribute]

            output += "\n".join(wrap(f"# {description}\n", subsequent_indent="# ")) + "\n"

            # Heuristic TOML dump
            if default_value is not None and default_value is not PydanticUndefined:
                try:
                    default_repr = json.dumps(default_value)
                except TypeError:
                    default_repr = json.dumps(str(default_value))

                output += f"{attribute} = {default_repr}\n"
            else:
                output += f"# {attribute} = ...\n"

            output += "\n"

        output += "\n"

    print(output.strip())


def main() -> None:
    gallia_commands = load_commands()
    parse_and_run(
        gallia_commands,
        top_level_options={
            "--version": (version, "show version and exit"),
            "--show-plugins": (show_plugins, "show registered plugins"),
            "--show-config": (show_config, "show loaded config"),
            "--template": (template, "generate a annotated config template"),
        },
        show_help_on_zero_args=True,
    )


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import asyncio
from collections.abc import Awaitable, Callable
from typing import Self

from gallia.log import get_logger
from gallia.power_supply.base import BasePowerSupplyDriver
from gallia.power_supply.devices.rnd import RND320
from gallia.power_supply.devices.rs.hmc804 import HMC804
from gallia.power_supply.uri import PowerSupplyURI

power_supply_drivers: dict[str, type[BasePowerSupplyDriver]] = {
    HMC804.PRODUCT_ID.lower(): HMC804,
    RND320.PRODUCT_ID.lower(): RND320,
}

logger = get_logger(__name__)


class PowerSupply:
    def __init__(self, driver: BasePowerSupplyDriver, channel_id: int | list[int]) -> None:
        self.channel_id = channel_id
        self.driver = driver
        self.mutex = asyncio.Lock()

    @classmethod
    async def connect(cls, target: PowerSupplyURI) -> Self:
        if target.product_id == "":
            raise ValueError("no device_id specified")

        driver = power_supply_drivers[target.product_id]
        client = await driver.connect(target, timeout=1.0)
        return cls(client, target.channel)

    async def _power(self, op: bool) -> None:
        assert self.driver
        if isinstance(self.channel_id, list):
            for id_ in self.channel_id:
                if id_ == 0:
                    await self.driver.set_master(op)
                else:
                    await self.driver.set_output(id_, op)
        elif isinstance(self.channel_id, int):
            if self.channel_id == 0:
                await self.driver.set_master(op)
            else:
                await self.driver.set_output(self.channel_id, op)

    async def power_up(self) -> None:
        logger.info("power up")
        await self._power(True)

    async def power_down(self) -> None:
        logger.info("power down")
        await self._power(False)

    async def power_cycle(
        self,
        sleep: float = 2.0,
        callback: Callable[[], Awaitable[None]] | None = None,
    ) -> None:
        async with self.mutex:
            await self.power_down()
            await asyncio.sleep(sleep)
            await self.power_up()
            if callback is not None:
                await callback()



# SPDX-FileCopyrightText: Hayden Richards
#
# SPDX-License-Identifier: MIT

"""Utilities to help with parsing arbitrarily nested `pydantic` models."""

from argparse import Namespace
from typing import Any, Generic, TypeAlias

from boltons.iterutils import get_path, remap  # type: ignore
from pydantic import BaseModel

from .namespaces import to_dict
from .pydantic import PydanticField, PydanticModelT

ModelT: TypeAlias = PydanticModelT | type[PydanticModelT] | BaseModel | type[BaseModel]


class _NestedArgumentParser(Generic[PydanticModelT]):
    """Parses arbitrarily nested `pydantic` models and inserts values passed at the command line."""

    def __init__(
        self,
        model: PydanticModelT | type[PydanticModelT],
        namespace: Namespace,
    ) -> None:
        self.model = model
        self.args = to_dict(namespace)
        self.subcommand_path: tuple[str, ...] = ()
        self.schema: dict[str, Any] = self._get_nested_model_fields(self.model, namespace)
        self.schema = self._remove_null_leaves(self.schema)

    def _get_nested_model_fields(self, model: ModelT[Any], namespace: Namespace) -> dict[str, Any]:
        def contains_subcommand(ns: Namespace, subcommand_path: tuple[str, ...]) -> bool:
            for step in subcommand_path:
                tmp = getattr(ns, step, None)

                if not isinstance(tmp, Namespace):
                    return False

                ns = tmp

            return True

        model_fields: dict[str, Any] = {}

        for field in PydanticField.parse_model(model):
            key = field.name

            if field.is_a(BaseModel):
                if field.is_subcommand():
                    sub_command_path = (*self.subcommand_path, key)

                    if not contains_subcommand(namespace, sub_command_path):
                        continue

                    self.subcommand_path = sub_command_path

                # recursively build nestes pydantic models in dict,
                # which matches the actual schema the nested
                # schema pydantic will be expecting
                model_fields[key] = self._get_nested_model_fields(field.model_type, namespace)
            else:
                # start with all leaves as None unless key is in top level
                value = self.args.get(key, None)

                if len(self.subcommand_path) > 0:
                    path = (*self.subcommand_path, key)
                    value = get_path(self.args, path, value)

                model_fields[key] = value

        return model_fields

    def _remove_null_leaves(self, schema: dict[str, Any]) -> Any:
        # only remove None leaves
        # actually CANNOT remove empty containers
        # this causes problems with nested submodels that don't
        # get any new args at command line, and therefore, are
        # relying on the submodel defaults
        # -> thus, the submodel name/key needs to be kept in
        # the schema
        return remap(schema, visit=lambda p, k, v: v is not None)

    def validate(self) -> tuple[PydanticModelT, BaseModel]:
        """Return the root of the model, as well as the sub-model for the bottom subcommand"""
        model = self.model.model_validate(self.schema)
        subcommand = model

        for step in self.subcommand_path:
            subcommand = getattr(subcommand, step)

        return model, subcommand



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import asyncio
import sys
from enum import IntEnum, unique
from typing import ClassVar, Self, TypeAlias

if sys.version_info < (3, 12):
    # NOTE: mypy's detection of these magic sys.foo checks is buggy
    #       and for some reason does not work here.
    from more_itertools import chunked as batched  # type: ignore[import-not-found]
else:
    from itertools import batched

from pydantic import BaseModel, ConfigDict, field_validator

from gallia.log import get_logger
from gallia.transports import BaseTransport, TargetURI, _ctypes_vector_xl, _ctypes_vector_xl_wrapper
from gallia.utils import auto_int

assert sys.platform == "win32", "unsupported platform"

logger = get_logger(__name__)


class RawFlexRayConfig(BaseModel):
    rx_queue_size: int = 0x20000
    channel_no: int | None = None

    @field_validator(
        "rx_queue_size",
        mode="before",
    )
    def auto_int(cls, v: str) -> int:
        return auto_int(v)


class FlexRayFrame(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    slot_id: int
    data: bytes
    raw: (
        _ctypes_vector_xl.XLfrEvent
        | _ctypes_vector_xl.XL_FR_RX_FRAME_EV
        | _ctypes_vector_xl.XL_FR_TX_FRAME_EV
        | None
    )


class RawFlexRayTransport(BaseTransport, scheme="fr-raw"):
    def __init__(self, target: TargetURI) -> None:
        super().__init__(target)

        self.check_scheme(target)
        self.config = RawFlexRayConfig.model_validate(target.qs_flat)

        self.backend = _ctypes_vector_xl_wrapper.FlexRayCtypesBackend.create(
            channel_no=self.config.channel_no,
            rx_queue_size=self.config.rx_queue_size,
        )
        self.backend.start_queue()

    def activate_channel(self) -> None:
        self.backend.activate_channel()

    def add_block_all_filter(self) -> None:
        self.backend.add_block_all_filter()

    def set_acceptance_filter(self, from_slot: int, to_slot: int | None = None) -> None:
        if to_slot is None:
            to_slot = from_slot

        self.backend.set_acceptance_filter(from_slot, to_slot)
        logger.debug(f"set accept filter: {from_slot}:{to_slot}")

    @classmethod
    async def connect(
        cls,
        target: str | TargetURI,
        timeout: float | None = None,
    ) -> Self:
        t = TargetURI(target) if isinstance(target, str) else target
        return cls(t)

    async def write_frame_unsafe(self, frame: FlexRayFrame) -> None:
        await self.backend.transmit(frame.slot_id, frame.data)
        logger.trace("wrote FlexRayFrame: %s", frame)

    async def write_frame(self, frame: FlexRayFrame) -> None:
        async with self.mutex:
            await self.write_frame_unsafe(frame)

    async def write(
        self,
        data: bytes,
        timeout: float | None = None,
        tags: list[str] | None = None,
    ) -> int:
        raise NotImplementedError

    async def read_frame_unsafe(
        self,
        slot_id: int | None = None,
        timeout: float | None = None,
        tags: list[str] | None = None,
    ) -> FlexRayFrame:
        event = await self.backend.receive(slot_id, timeout)

        frame = FlexRayFrame(
            slot_id=event.slotID,
            data=bytes(event.data)[: int(event.payloadLength)],
            raw=event,
        )

        logger.trace("read FlexRayFrame: %s", frame)
        return frame

    async def read_frame(
        self,
        slot_id: int | None = None,
        timeout: float | None = None,
        tags: list[str] | None = None,
    ) -> FlexRayFrame:
        async with self.mutex:
            return await self.read_frame_unsafe(slot_id, timeout, tags)

    async def read(self, timeout: float | None = None, tags: list[str] | None = None) -> bytes:
        raise NotImplementedError

    async def close(self) -> None:
        async with self.mutex:
            await self.backend.close()


class FlexrayTPLegacyConfig(BaseModel):
    src_slot_id: int
    dst_slot_id: int
    src_address: int
    dst_address: int
    payload_rx_start_index: int
    payload_rx_end_index: int
    fc_block_size: int = 0xFF
    fc_separation_time: int = 0x04
    tx_sleep_time: float = 0.05

    @field_validator(
        "src_slot_id",
        "dst_slot_id",
        "src_address",
        "dst_address",
        "payload_rx_start_index",
        "payload_rx_end_index",
        "fc_block_size",
        "fc_separation_time",
        mode="before",
    )
    def auto_int(cls, v: str) -> int:
        return auto_int(v)


@unique
class FlexRayTPFrameType(IntEnum):
    SINGLE_FRAME = 0x00
    FIRST_FRAME = 0x01
    CONSECUTIVE_FRAME = 0x02
    FLOW_CONTROL_FRAME = 0x03


@unique
class FlexRayTPFlowControlFlag(IntEnum):
    CONTINUE_TO_SEND = 0
    WAIT = 1
    ABORT = 2


def parse_frame_type(data: bytes) -> FlexRayTPFrameType:
    return FlexRayTPFrameType(data[0] >> 4)


class FlexRayTPSingleFrame(BaseModel):
    type_: ClassVar[FlexRayTPFrameType] = FlexRayTPFrameType.SINGLE_FRAME
    data: bytes
    size: int

    def __bytes__(self) -> bytes:
        return self.frame_header + self.data

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} size: {self.size} data: {self.data.hex()}>"

    @property
    def frame_header(self) -> bytes:
        return bytes([((self.type_ << 4) & 0xF0) | self.size & 0x0F])

    @classmethod
    def parse(cls, data: bytes) -> Self:
        type_ = parse_frame_type(data)
        if type_ != cls.type_:
            raise ValueError(f"wrong frame type: {type:x}")
        size = data[0] & 0xF
        return cls(data=data[1 : size + 1], size=size)


class FlexRayTPFirstFrame(BaseModel):
    type_: ClassVar[FlexRayTPFrameType] = FlexRayTPFrameType.FIRST_FRAME
    data: bytes
    size: int

    def __bytes__(self) -> bytes:
        return self.frame_header + self.data

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} size: {self.size} data: {self.data.hex()}>"

    @property
    def frame_header(self) -> bytes:
        return bytes([(((self.type_ << 4) & 0xF0) | (self.size >> 8) & 0x0F), self.size & 0xFF])

    @classmethod
    def parse(cls, data: bytes) -> Self:
        type_ = parse_frame_type(data)
        if type_ != cls.type_:
            raise ValueError(f"wrong frame type_: {type:x}")

        size = ((data[0] & 0x0F) << 8) | data[1]
        return cls(data=data[2 : size + 1], size=size)


class FlexRayTPConsecutiveFrame(BaseModel):
    type_: ClassVar[FlexRayTPFrameType] = FlexRayTPFrameType.CONSECUTIVE_FRAME
    counter: int
    data: bytes

    def __bytes__(self) -> bytes:
        return self.frame_header + self.data

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} counter: {self.counter} data: {self.data.hex()}>"

    @property
    def frame_header(self) -> bytes:
        return bytes([((self.type_ << 4) & 0xF0) | self.counter & 0x0F])

    @classmethod
    def parse(cls, data: bytes) -> Self:
        type_ = parse_frame_type(data)
        if type_ != cls.type_:
            raise ValueError(f"wrong frame type_: {type:x}")
        counter = data[0] & 0xF
        return cls(counter=counter, data=data[1:])


class FlexRayTPFlowControlFrame(BaseModel):
    type_: ClassVar[FlexRayTPFrameType] = FlexRayTPFrameType.FLOW_CONTROL_FRAME
    flag: FlexRayTPFlowControlFlag
    block_size: int
    separation_time: int

    def __bytes__(self) -> bytes:
        return bytes(
            [
                ((self.type_ << 4) & 0xF0) | (self.flag & 0x0F),
                self.block_size,
                self.separation_time,
            ]
        )

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} flag: {self.flag} block_size: {self.block_size} separation_time: {self.separation_time}>"

    @classmethod
    def parse(cls, data: bytes) -> Self:
        type_ = FlexRayTPFrameType(data[0] >> 4)
        if type_ != cls.type_:
            raise ValueError(f"wrong frame type_: {type:x}")
        flag = FlexRayTPFlowControlFlag(data[0] & 0x0F)
        block_size = data[1]
        separation_time = data[2]
        return cls(
            flag=flag,
            block_size=block_size,
            separation_time=separation_time,
        )


FlexRayTPFrame: TypeAlias = (
    FlexRayTPSingleFrame
    | FlexRayTPFirstFrame
    | FlexRayTPConsecutiveFrame
    | FlexRayTPFlowControlFrame
)


def parse_frame(data: bytes) -> FlexRayTPFrame:
    t = parse_frame_type(data)
    match t:
        case FlexRayTPFrameType.SINGLE_FRAME:
            return FlexRayTPSingleFrame.parse(data)
        case FlexRayTPFrameType.FIRST_FRAME:
            return FlexRayTPFirstFrame.parse(data)
        case FlexRayTPFrameType.CONSECUTIVE_FRAME:
            return FlexRayTPConsecutiveFrame.parse(data)
        case FlexRayTPFrameType.FLOW_CONTROL_FRAME:
            return FlexRayTPFlowControlFrame.parse(data)


class FlexRayTPLegacyTransport(BaseTransport, scheme="fr-tp-legacy"):
    def __init__(self, target: TargetURI, fr_raw: RawFlexRayTransport) -> None:
        super().__init__(target)

        self.check_scheme(target)
        self.config = FlexrayTPLegacyConfig.model_validate(target.qs_flat)
        self.mutex = asyncio.Lock()

        self.fr_raw = fr_raw
        self.fr_raw.add_block_all_filter()
        self.fr_raw.set_acceptance_filter(self.config.src_slot_id, self.config.src_slot_id)
        self.fr_raw.set_acceptance_filter(self.config.dst_slot_id, self.config.dst_slot_id)
        self.fr_raw.activate_channel()

    @classmethod
    async def connect(
        cls,
        target: str | TargetURI,
        timeout: float | None = None,
    ) -> Self:
        t = TargetURI(target) if isinstance(target, str) else target
        fr_raw = await RawFlexRayTransport.connect("fr-raw:", timeout)
        return cls(t, fr_raw)

    async def write_bytes(self, data: bytes) -> None:
        frame = FlexRayFrame(
            data=data,
            slot_id=self.config.dst_slot_id,
            raw=None,  # raw is constructed by the layers underneath
        )
        await self.fr_raw.write_frame(frame)

    async def write_tp_frame(self, frame: FlexRayTPFrame) -> None:
        logger.trace("write FlexRayTPFrame: %s", frame)
        address_header = bytes(
            [
                ((self.config.dst_address >> 8) & 0xFF),
                (self.config.dst_address & 0xFF),
                ((self.config.src_address >> 8) & 0xFF),
                (self.config.src_address & 0xFF),
            ],
        )
        await self.write_bytes(address_header + bytes(frame))

    async def write_unsafe(
        self,
        data: bytes,
        timeout: float | None = None,
        tags: list[str] | None = None,
    ) -> int:
        # Single Frame is sufficiant.
        if len(data) < 8:
            await self.write_tp_frame(FlexRayTPSingleFrame(data=data, size=len(data)))
            return len(data)

        # Write a first frame…
        first_frame = FlexRayTPFirstFrame(data=data[:6], size=len(data))
        await self.write_tp_frame(first_frame)

        # … then a flow control comes.
        fc_frame = await self.read_tp_frame()
        if not isinstance(fc_frame, FlexRayTPFlowControlFrame):
            raise RuntimeError(f"unexpected frame received: {fc_frame}")

        # Best effort, just send the data.
        # TODO: Not implemented: block size handling.
        # Maybe a further flow control comes after block size number of frames.

        counter = 1
        for batch in batched(data[6:], 7):
            cf_frame = FlexRayTPConsecutiveFrame(
                counter=counter,
                data=bytes(batch),
            )

            await self.write_tp_frame(cf_frame)

            # XXX: FlexRay has a defined schedule.
            # In the best case, we get notified that our
            # frame has been written to the bus. We do
            # not have this information, so we sleep for
            # a configurable time frame…
            await asyncio.sleep(self.config.tx_sleep_time)
            counter = (counter + 1) & 0x0F

        logger.debug("wrote data: %s", data.hex())
        return len(data)

    async def write(
        self,
        data: bytes,
        timeout: float | None = None,
        tags: list[str] | None = None,
    ) -> int:
        async with self.mutex:
            await self.write_unsafe(data, timeout, tags)
            return len(data)

    async def read_bytes(self) -> bytes:
        while True:
            frame = await self.fr_raw.read_frame(self.config.src_slot_id)
            if frame.data[0] == 0x00:
                continue
            return frame.data[self.config.payload_rx_start_index : self.config.payload_rx_end_index]

    @staticmethod
    def _parse_address(data: bytes) -> tuple[int, int]:
        dst_address = data[0] << 8 | data[1]
        src_address = data[2] << 8 | data[3]
        return dst_address, src_address

    async def read_tp_frame(self) -> FlexRayTPFrame:
        data = await self.read_bytes()
        dst_address, src_address = self._parse_address(data)
        # logger.trace("got frame for addresses: %x %x", dst_address, src_address)
        frame = parse_frame(data[4:])
        logger.trace("read FlexRayTPFrame %s", repr(frame))
        return frame

    def _require_fc_frame(self, block_size: int, read_bytes: int) -> bool:
        # 6 bytes already read in first frame.
        return ((read_bytes - 6) & block_size) == 0

    async def _send_flow_control_frame(self) -> None:
        block_size = self.config.fc_block_size
        fc_frame = FlexRayTPFlowControlFrame(
            flag=FlexRayTPFlowControlFlag.CONTINUE_TO_SEND,
            separation_time=self.config.fc_separation_time,
            block_size=block_size,
        )
        await self.write_tp_frame(fc_frame)

    async def _handle_fragmented(self, expected_len: int) -> bytes:
        # 6 bytes already read in first frame.
        # Headersize is 2 byte.
        read_bytes = 6
        counter = 1
        data = b""

        while read_bytes < expected_len:
            # Reordering is not implemented.
            logger.debug(f"expected_len: {expected_len}; read_bytes: {read_bytes}")

            if self._require_fc_frame(self.config.fc_block_size, read_bytes):
                await self._send_flow_control_frame()

            # TODO: Make this configurable. Maybe align with separation_time.
            async with asyncio.timeout(10):
                frame = await self.read_tp_frame()

            if not isinstance(frame, FlexRayTPConsecutiveFrame):
                raise RuntimeError(f"expected consecutive frame, got: {frame}")
            if frame.counter != (counter & 0x0F):
                raise RuntimeError(f"got unexpected consecutive counter: {frame.counter}")

            # Header size needs to be respected here.
            read_bytes += len(frame.data)
            data += frame.data
            counter = (counter + 1) & 0x0F

        return data

    async def read_unsafe(
        self,
        timeout: float | None = None,
        tags: list[str] | None = None,
    ) -> bytes:
        async with asyncio.timeout(timeout):
            frame = await self.read_tp_frame()
            match frame:
                case FlexRayTPSingleFrame():
                    return frame.data
                case FlexRayTPFirstFrame():
                    data = frame.data + await self._handle_fragmented(frame.size)
                    data = data[: frame.size]
                    logger.debug("read data: %s", data.hex())
                    return data
                case _:
                    raise RuntimeError(f"got unexpected tp frame: {frame}")

    async def read(
        self,
        timeout: float | None = None,
        tags: list[str] | None = None,
    ) -> bytes:
        async with self.mutex:
            return await self.read_unsafe(timeout, tags)

    async def close(self) -> None:
        await self.fr_raw.close()



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import asyncio
import sys
from typing import Self

from pydantic import model_validator

from gallia.command import UDSScanner
from gallia.command.config import AutoInt, Field, HexBytes
from gallia.command.uds import UDSScannerConfig
from gallia.log import get_logger
from gallia.services.uds import NegativeResponse
from gallia.services.uds.core.service import RoutineControlResponse
from gallia.services.uds.core.utils import g_repr

logger = get_logger(__name__)


class RTCLPrimitiveConfig(UDSScannerConfig):
    properties: bool = Field(
        False,
        description="Read and store the ECU proporties prior and after scan",
        cli_group=UDSScannerConfig._cli_group,
        config_section=UDSScannerConfig._config_section,
    )
    session: AutoInt | None = Field(None, description="The session in which the requests are made")
    routine_identifier: AutoInt = Field(description="The routine identifier", positional=True)
    start: bool = Field(
        False,
        description="Start the routine with a startRoutine request (this task is always executed first)",
    )
    stop: bool = Field(
        False,
        description="Stop the routine with a stopRoutine request (this task is executed after starting the routine if --start is given as well)",
    )
    results: bool = Field(
        False,
        description="Read the routine results with a requestRoutineResults request (this task is always executed last)",
    )
    start_parameters: HexBytes = Field(
        b"",
        description="The routineControlOptionRecord passed to the startRoutine request",
        metavar="HEXSTRING",
    )
    stop_parameters: HexBytes = Field(
        b"",
        description="The routineControlOptionRecord passed to the stopRoutine request",
        metavar="HEXSTRING",
    )
    results_parameters: HexBytes = Field(
        b"",
        description="The routineControlOptionRecord passed to the stopRoutine request",
        metavar="HEXSTRING",
    )
    stop_delay: float = Field(
        0.0,
        description="Delay the stopRoutine request by the given amount of seconds",
        metavar="SECONDS",
    )
    results_delay: float = Field(
        0.0,
        description="Delay the requestRoutineResults request by the given amount of seconds",
        metavar="SECONDS",
    )

    @model_validator(mode="after")
    def check_action(self) -> Self:
        if not any((self.start, self.stop, self.results)):
            raise ValueError("No instructions were given (start/stop/results)")

        return self


class RTCLPrimitive(UDSScanner):
    """Start or stop a provided routine or request its results"""

    CONFIG_TYPE = RTCLPrimitiveConfig
    SHORT_HELP = "RoutineControl"

    def __init__(self, config: RTCLPrimitiveConfig):
        super().__init__(config)
        self.config: RTCLPrimitiveConfig = config

    async def main(self) -> None:
        if self.config.session is not None:
            try:
                await self.ecu.check_and_set_session(self.config.session)
            except Exception as e:
                logger.critical(
                    f"Could not change to session: {g_repr(self.config.session)}: {e!r}"
                )
                sys.exit(1)

        if self.config.start:
            resp: (
                NegativeResponse | RoutineControlResponse
            ) = await self.ecu.routine_control_start_routine(
                self.config.routine_identifier, self.config.start_parameters
            )

            if isinstance(resp, NegativeResponse):
                logger.error(f"start_routine: {resp}")
            else:
                logger.result("[start] Positive response:")
                logger.result(f"hex: {resp.routine_status_record.hex()}")
                logger.result(f"raw: {resp.routine_status_record!r}")

        if self.config.stop:
            delay = self.config.stop_delay

            if delay > 0:
                logger.info(f"Delaying the request for stopping the routine by {delay} seconds")
                await asyncio.sleep(delay)

            resp = await self.ecu.routine_control_stop_routine(
                self.config.routine_identifier, self.config.stop_parameters
            )

            if isinstance(resp, NegativeResponse):
                logger.error(f"stop routine: {resp}")
            else:
                logger.result("[stop] Positive response:")
                logger.result(f"hex: {resp.routine_status_record.hex()}")
                logger.result(f"raw: {resp.routine_status_record!r}")

        if self.config.results:
            delay = self.config.results_delay

            if delay > 0:
                logger.info(f"Delaying the request for the routine results by {delay} seconds")
                await asyncio.sleep(delay)

            resp = await self.ecu.routine_control_request_routine_results(
                self.config.routine_identifier, self.config.results_parameters
            )

            if isinstance(resp, NegativeResponse):
                logger.error(f"request_routine_results: {resp}")
            else:
                logger.result("[results] Positive response:")
                logger.result(f"hex: {resp.routine_status_record.hex()}")
                logger.result(f"raw: {resp.routine_status_record!r}")



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0


from gallia.command import Scanner
from gallia.command.base import ScannerConfig
from gallia.command.config import Field, HexBytes
from gallia.command.uds import UDSScannerConfig


class GenericPDUPrimitiveConfig(ScannerConfig):
    properties: bool = Field(
        False,
        description="Read and store the ECU proporties prior and after scan",
        cli_group=UDSScannerConfig._cli_group,
        config_section=UDSScannerConfig._config_section,
    )
    pdu: HexBytes = Field(description="raw pdu to send", positional=True)


class GenericPDUPrimitive(Scanner):
    """A raw scanner to send a plain pdu"""

    CONFIG_TYPE = GenericPDUPrimitiveConfig
    SHORT_HELP = "send a plain PDU"

    def __init__(self, config: GenericPDUPrimitiveConfig):
        super().__init__(config)
        self.config: GenericPDUPrimitiveConfig = config

    async def main(self) -> None:
        await self.transport.write(self.config.pdu)



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import asyncio
from pathlib import Path
from typing import Any

from gallia.power_supply.base import BasePowerSupplyDriver
from gallia.power_supply.exceptions import OperationNotSupportedError


class RND320(BasePowerSupplyDriver):
    PRODUCT_ID = "RND320"

    def _send(self, data: str) -> None:
        with Path(self.target.path).open("w") as f:
            f.write(data)

    async def get_ident(self) -> str:
        # This endpoint is completely broken in the device.
        # Return a static string to satisfy gallia's driver
        # interface.
        return self.PRODUCT_ID

    async def get_master(self) -> bool:
        raise OperationNotSupportedError

    async def set_master(self, enabled: bool) -> None:
        cmd = "OUT1" if enabled else "OUT0"
        await asyncio.to_thread(self._send, cmd)

    async def get_channels(self) -> int:
        return 1

    async def get_current(self, channel: int) -> float:
        raise OperationNotSupportedError

    async def set_current(self, channel: int, value: float) -> None:
        raise OperationNotSupportedError

    async def get_voltage(self, channel: int) -> float:
        raise OperationNotSupportedError

    async def set_voltage(self, channel: int, value: float) -> None:
        raise OperationNotSupportedError

    async def get_output(self, channel: int) -> bool:
        raise OperationNotSupportedError

    async def set_output(self, channel: int, enabled: bool) -> None:
        await self.set_master(enabled)

    async def status(self) -> dict[str, Any]:
        raise OperationNotSupportedError

    async def get_ocp(self, channel: int) -> bool:
        raise OperationNotSupportedError

    async def set_ocp(self, channel: int, enabled: bool) -> None:
        raise OperationNotSupportedError

    async def get_ovp(self, channel: int) -> bool:
        raise OperationNotSupportedError

    async def set_ovp(self, channel: int, enabled: bool) -> None:
        raise OperationNotSupportedError

    async def set_beep(self, enabled: bool) -> None:
        raise OperationNotSupportedError



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import sys
from pathlib import Path
from typing import Self

from pydantic import model_validator

from gallia.command import UDSScanner
from gallia.command.config import AutoInt, Field, HexBytes
from gallia.command.uds import UDSScannerConfig
from gallia.log import get_logger
from gallia.services.uds import NegativeResponse
from gallia.services.uds.core.utils import g_repr

logger = get_logger(__name__)


class WMBAPrimitiveConfig(UDSScannerConfig):
    session: AutoInt = Field(0x01, description="The session in which the requests are made")
    address: AutoInt = Field(
        description="The start address to which data should be written", positional=True
    )
    data: HexBytes | None = Field(None, description="The data which should be written")
    data_file: Path | None = Field(
        None, description="The path to a file with the binary data which should be written"
    )

    @model_validator(mode="after")
    def check_data_source(self) -> Self:
        if not (self.data is None) ^ (self.data_file is None):
            raise ValueError("Exactly one of data or data-file is required")

        return self


class WMBAPrimitive(UDSScanner):
    """Write memory by address"""

    CONFIG_TYPE = WMBAPrimitiveConfig
    SHORT_HELP = "WriteMemoryByAddress"

    def __init__(self, config: WMBAPrimitiveConfig):
        super().__init__(config)
        self.config: WMBAPrimitiveConfig = config

    async def main(self) -> None:
        try:
            await self.ecu.check_and_set_session(self.config.session)
        except Exception as e:
            logger.critical(f"Could not change to session: {g_repr(self.config.session)}: {e!r}")
            sys.exit(1)

        if self.config.data is not None:
            data = self.config.data
        else:
            assert self.config.data_file is not None

            with self.config.data_file.open("rb") as file:
                data = file.read()

        resp = await self.ecu.write_memory_by_address(self.config.address, data)

        if isinstance(resp, NegativeResponse):
            logger.error(resp)
        else:
            # There is not real data returned, only echos
            logger.result("Success")



# SPDX-FileCopyrightText: Hayden Richards
#
# SPDX-License-Identifier: MIT

"""Declarative Typed Argument Parsing with Pydantic Models.

This is the `pydantic-argparse` package, which contains the classes, methods
and functions required for declarative and typed argument parsing with
`pydantic` models.

The public interface exposed by this package is the declarative and typed
`ArgumentParser` class, as well as the package "dunder" metadata.
"""

# Local
from pydantic import BaseModel, ConfigDict

from gallia.pydantic_argparse.argparse import ArgumentParser


class BaseArgument(BaseModel):
    """Base pydantic model for argument groups."""

    model_config = ConfigDict(json_schema_extra={"subcommand": False})


class BaseCommand(BaseModel):
    """Base pydantic model for command groups.

    This class is only a convenience base class that sets the
    `model_config` parameter to have the `json_schema_extra` parameter to
    have `subcommand=True`.
    """

    model_config = ConfigDict(json_schema_extra={"subcommand": True}, defer_build=True)


# Public Re-Exports
__all__ = (
    "ArgumentParser",
    "BaseArgument",
    "BaseCommand",
)



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import sys
from collections.abc import Mapping

from gallia.command import BaseCommand
from gallia.plugins.plugin import CommandTree, Plugin


class XCPPlugin(Plugin):
    @classmethod
    def name(cls) -> str:
        return "Gallia XCP"

    @classmethod
    def description(cls) -> str:
        return "Default Gallia plugin for Universal Measurement and Calibration Protocol (XCP) functionality"

    @classmethod
    def commands(cls) -> Mapping[str, CommandTree | type[BaseCommand]]:
        tree: dict[str, CommandTree | type[BaseCommand]] = {}

        if sys.platform.startswith("linux"):
            from gallia.commands.discover.find_xcp import CanFindXCP, TcpFindXCP, UdpFindXCP
            from gallia.commands.primitive.uds.xcp import SimpleTestXCP

            tree = {
                "discover": CommandTree(
                    description=None,
                    subtree={
                        "xcp": CommandTree(
                            description="XCP enumeration scanner",
                            subtree={
                                "can": CanFindXCP,
                                "tcp": TcpFindXCP,
                                "udp": UdpFindXCP,
                            },
                        ),
                    },
                ),
                "primitive": CommandTree(
                    description=None,
                    subtree={
                        "xcp": SimpleTestXCP,
                    },
                ),
            }

        return tree



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import sys
from enum import StrEnum, unique

TCP = "tcp"
TCP_LINES = "tcp-lines"
DOIP = "doip"
DUMMY = "dummy"


if sys.platform.startswith("linux"):

    @unique
    class TransportScheme(StrEnum):
        TCP = TCP
        TCP_LINES = TCP_LINES
        DOIP = DOIP
        DUMMY = DUMMY

        HSFZ = "hsfz"
        UNIX = "unix"
        UNIX_LINES = "unix-lines"
        ISOTP = "isotp"
        CAN_RAW = "can-raw"


if sys.platform == "win32":

    @unique
    class TransportScheme(StrEnum):
        TCP = TCP
        TCP_LINES = TCP_LINES
        DOIP = DOIP
        DUMMY = DUMMY

        FLEXRAY_RAW = "fr-raw"
        FLEXRAY_TP_LEGACY = "fr-tp-legacy"



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import sys
from typing import Self

from pydantic import model_validator

assert sys.platform.startswith("linux"), "unsupported platform"

from gallia.command import Scanner
from gallia.command.base import ScannerConfig
from gallia.command.config import AutoInt, Field
from gallia.plugins.plugin import load_transport
from gallia.services.xcp import CANXCPSerivce, XCPService
from gallia.transports import ISOTPTransport, RawCANTransport
from gallia.utils import catch_and_log_exception


class SimpleTestXCPConfig(ScannerConfig):
    can_master: AutoInt | None = Field(None)
    can_slave: AutoInt | None = Field(None)

    @model_validator(mode="after")
    def check_transport_requirements(self) -> Self:
        if self.target.scheme == RawCANTransport.SCHEME and (
            self.can_master is None or self.can_slave is None
        ):
            raise ValueError("For CAN interfaces, master and slave address are required!")

        if self.target.scheme == ISOTPTransport.SCHEME:
            raise ValueError("Use can-raw for CAN interfaces!")

        return self


class SimpleTestXCP(Scanner):
    """Test XCP Slave"""

    CONFIG_TYPE = SimpleTestXCPConfig
    SHORT_HELP = "XCP tester"

    def __init__(self, config: SimpleTestXCPConfig):
        super().__init__(config)
        self.config: SimpleTestXCPConfig = config
        self.service: XCPService

    async def setup(self) -> None:
        transport_type = load_transport(self.config.target)
        transport = await transport_type.connect(self.config.target)

        if isinstance(transport, RawCANTransport):
            assert self.config.can_master is not None and self.config.can_slave is not None

            self.service = CANXCPSerivce(transport, self.config.can_master, self.config.can_slave)
        else:
            self.service = XCPService(transport)

        await super().setup()

    async def main(self) -> None:
        await catch_and_log_exception(self.service.connect)
        await catch_and_log_exception(self.service.get_status)
        await catch_and_log_exception(self.service.get_comm_mode_info)
        await catch_and_log_exception(self.service.disconnect)



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0
#
# Preferred system exit codes as defined by sysexits.h
#
# Exit code constants intended to be passed to `sys.exit()`;
# inspired by https://docs.rs/exitcode.

# Copy and pasted from: https://github.com/rumpelsepp/exitcode/blob/176eaaf8b70905fe18a904670b481f7a00edfc68/exitcode/__init__.py

# Successful exit
OK = 0

# The command was used incorrectly, e.g., with the
# wrong number of arguments, a bad flag, a bad syntax
# in a parameter, etc.
USAGE = 64

# The input data was incorrect in some way.  This
# should only be used for user's data and not system
# files.
DATAERR = 65

# An input file (not a system file) did not exist or
# was not readable.  This could also include errors
# like "No message" to a mailer (if it cared to
# catch it).
NOINPUT = 66

# The user specified did not exist.  This might be
# used for mail addresses or remote logins.
NOUSER = 67

# The host specified did not exist.  This is used in
# mail addresses or network requests.
NOHOST = 68

# A service is unavailable.  This can occur if a
# support program or file does not exist. This can also
# be used as a catchall message when something you
# wanted to do doesn't work, but you don't know why.
UNAVAILABLE = 69

# An internal software error has been detected.  This
# should be limited to non-operating system related
# errors as possible.
SOFTWARE = 70

# An operating system error has been detected.  This
# is intended to be used for such things as "cannot
# fork", "cannot create pipe", or the like.  It
# includes things like getuid returning a user that
# does not exist in the passwd file.
OSERR = 71

# Some system file (e.g., /etc/passwd, /var/run/utmp,
# etc.) does not exist, cannot be opened, or has some
# sort of error (e.g., syntax error).
OSFILE = 72

# A (user specified) output file cannot be created.
CANTCREAT = 73

# An error occurred while doing I/O on some file.
IOERR = 74

# Temporary failure, indicating something that is not
# really an error.  In sendmail, this means that a
# mailer (e.g.) could not create a connection, and
# the request should be reattempted later.
TEMPFAIL = 75

# The remote system returned something that was
# "not possible" during a protocol exchange.
PROTOCOL = 76

# You did not have sufficient permission to perform
# the operation.  This is not intended for file system
# problems, which should use `NOINPUT` or `CANTCREAT`,
# but rather for higher level permissions.
NOPERM = 77

# Something was found in an unconfigured or misconfigured state.
CONFIG = 78



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import sys
from typing import Literal

from gallia.command import UDSScanner
from gallia.command.config import AutoInt, Field, HexBytes
from gallia.command.uds import UDSScannerConfig
from gallia.log import get_logger
from gallia.services.uds import NegativeResponse
from gallia.services.uds.core.utils import g_repr

logger = get_logger(__name__)


class IOCBIPrimitiveConfig(UDSScannerConfig):
    properties: bool = Field(
        False,
        description="Read and store the ECU proporties prior and after scan",
        cli_group=UDSScannerConfig._cli_group,
        config_section=UDSScannerConfig._config_section,
    )
    session: AutoInt = Field(0x01, description="The session in which the requests are made")
    data_identifier: AutoInt = Field(description="The data identifier", positional=True)
    control_parameter: Literal[
        "return-control-to-ecu",
        "reset-to-default",
        "freeze-current-state",
        "short-term-adjustment",
        "without-control-parameter",
    ] = Field(
        description='Control parameter sent to the ECU. "short-term-adjustment" and "without-control-parameter" require passing a new state as well.',
        positional=True,
    )
    new_state: HexBytes = Field(
        b"",
        description='The new state required in use with the two control parameters "short-term-adjustment" and "without-control-parameter".',
        metavar="HEXSTRING",
    )
    control_enable_mask: HexBytes = Field(
        b"",
        description="This parameter is used if the data-identifier corresponds to multiple signals.In that case each bit enables or disables setting of each corresponding signal.Can only be used in combination with a control parameter.",
        metavar="HEXSTRING",
    )


class IOCBIPrimitive(UDSScanner):
    """Input output control"""

    CONFIG_TYPE = IOCBIPrimitiveConfig
    SHORT_HELP = "InputOutputControl"

    def __init__(self, config: IOCBIPrimitiveConfig):
        super().__init__(config)
        self.config: IOCBIPrimitiveConfig = config

    async def main(self) -> None:
        try:
            await self.ecu.check_and_set_session(self.config.session)
        except Exception as e:
            logger.critical(f"Could not change to session: {g_repr(self.config.session)}: {e!r}")
            sys.exit(1)

        did = self.config.data_identifier
        control_enable_mask_record = self.config.control_enable_mask
        uses_control_parameter = True

        if self.config.control_parameter == "return-control-to-ecu":
            resp = await self.ecu.input_output_control_by_identifier_return_control_to_ecu(
                did, control_enable_mask_record
            )
        elif self.config.control_parameter == "reset-to-default":
            resp = await self.ecu.input_output_control_by_identifier_reset_to_default(
                did, control_enable_mask_record
            )
        elif self.config.control_parameter == "freeze-current-state":
            resp = await self.ecu.input_output_control_by_identifier_freeze_current_state(
                did, control_enable_mask_record
            )
        elif self.config.control_parameter == "short-term-adjustment":
            resp = await self.ecu.input_output_control_by_identifier_short_term_adjustment(
                did, self.config.new_state, control_enable_mask_record
            )
        elif self.config.control_parameter == "without-control-parameter":
            resp = await self.ecu.input_output_control_by_identifier(
                did, self.config.new_state, control_enable_mask_record
            )
            uses_control_parameter = False

        if isinstance(resp, NegativeResponse):
            logger.error(resp)
        else:
            logger.result("Positive response:")
            data = (
                resp.control_status_record[1:]
                if uses_control_parameter
                else resp.control_status_record
            )
            logger.result(f"hex: {data.hex()}")
            logger.result(f"raw: {repr(data)}")



# SPDX-FileCopyrightText: Hayden Richards
#
# SPDX-License-Identifier: MIT

"""Declarative and Typed Argument Parser.

The `parser` module contains the `ArgumentParser` class, which provides a
declarative method of defining command-line interfaces.

The procedure to declaratively define a typed command-line interface is:

1. Define `pydantic` arguments model
2. Create typed `ArgumentParser`
3. Parse typed arguments

The resultant arguments object returned is an instance of the defined
`pydantic` model. This means that the arguments object and its attributes will
be compatible with an IDE, linter or type checker.
"""

import argparse
import sys
from typing import Any, Generic, Never, NoReturn

from pydantic import BaseModel, ValidationError

from gallia.pydantic_argparse import parsers
from gallia.pydantic_argparse.argparse import actions
from gallia.pydantic_argparse.parsers import command
from gallia.pydantic_argparse.utils.field import ArgFieldInfo
from gallia.pydantic_argparse.utils.nesting import _NestedArgumentParser
from gallia.pydantic_argparse.utils.pydantic import PydanticField, PydanticModelT


class ArgumentParser(argparse.ArgumentParser, Generic[PydanticModelT]):
    """Declarative and Typed Argument Parser.

    The `ArgumentParser` declaratively generates a command-line interface using
    the `pydantic` model specified upon instantiation.

    The `ArgumentParser` provides the following `argparse` functionality:

    * Argument Groups
    * Subcommands

    All arguments are *named*, and positional arguments are not supported.

    The `ArgumentParser` provides the method `parse_typed_args()` to parse
    command line arguments and return an instance of its bound `pydantic`
    model, populated with the parsed and validated user supplied command-line
    arguments.
    """

    # Argument Group Names
    COMMANDS = "commands"
    HELP = "help"

    # Exit Codes
    EXIT_ERROR = 2

    def __init__(
        self,
        model: type[PydanticModelT],
        prog: str | None = None,
        description: str | None = None,
        version: str | None = None,
        epilog: str | None = None,
        add_help: bool = True,
        exit_on_error: bool = True,
        extra_defaults: dict[type, dict[str, tuple[str, Any]]] | None = None,
    ) -> None:
        """Instantiates the typed Argument Parser with its `pydantic` model.

        :param model: Pydantic argument model class.
        :param prog: Program name for CLI.
        :param description: Program description for CLI.
        :param version: Program version string for CLI.
        :param epilog: Optional text following help message.
        :param add_help: Whether to add a `-h`/`--help` flag.
        :param exit_on_error: Whether to exit on error.
        :param extra_defaults: Defaults coming from external sources, such as environment variables or config files.
        """
        # Initialise Super Class
        super().__init__(
            prog=prog,
            description=description,
            epilog=epilog,
            exit_on_error=exit_on_error,
            add_help=False,  # Always disable the automatic help flag.
            argument_default=argparse.SUPPRESS,  # Allow `pydantic` to handle defaults.
            formatter_class=argparse.RawTextHelpFormatter,
        )

        # Set Version, Add Help and Exit on Error Flag
        self.version = version
        self.add_help = add_help
        self.exit_on_error = exit_on_error
        self.extra_defaults = extra_defaults

        # Add Arguments Groups
        self._subcommands: argparse._SubParsersAction[Any] | None = None

        # Add Arguments from Model
        self._submodels: dict[str, type[BaseModel]] = {}
        self.model = model
        self._add_model(model)

        self._help_group = self.add_argument_group(ArgumentParser.HELP)

        # Add Help and Version Flags
        if self.add_help:
            self._add_help_flag()
        if self.version:
            self._add_version_flag()

    def parse_typed_args(
        self,
        args: list[str] | None = None,
    ) -> tuple[PydanticModelT, BaseModel]:
        """Parses command line arguments.

        If `args` are not supplied by the user, then they are automatically
        retrieved from the `sys.argv` command-line arguments.

        :param args: Optional list of arguments to parse.
        :return: A tuple of the whole parsed model, as well as the submodel representing the selected subcommand.
        """
        # Call Super Class Method
        namespace = self.parse_args(args)
        nested_parser = _NestedArgumentParser(model=self.model, namespace=namespace)

        try:
            return nested_parser.validate()
        except ValidationError as exc:
            # Catch exceptions, and use the ArgumentParser.error() method
            # to report it to the user
            self._validation_error(exc, nested_parser)

    def _validation_error(
        self, error: ValidationError, parser: _NestedArgumentParser[Any]
    ) -> Never:
        self.print_usage(sys.stderr)

        model = parser.model
        for scp in parser.subcommand_path:
            model = PydanticField(scp, model.model_fields[scp]).model_type

        fields = model.model_fields
        msg = ""

        if error.error_count() == 1:
            msg += "error: "
        else:
            msg += f"{error.error_count()} errors: \n"

        for e in error.errors():
            if error.error_count() > 1:
                msg += "  "

            source = ""
            sources = e["loc"][len(parser.subcommand_path) :]

            # If the validation failed for a field validator there is one source level left,
            # which equals the name of the field
            if len(sources) > 0:
                argument = sources[0]

                assert isinstance(argument, str)

                if (
                    self.extra_defaults is not None
                    and model in self.extra_defaults
                    and argument in self.extra_defaults[model]
                    and self.extra_defaults[model][argument][1] == e["input"]
                ):
                    source = (
                        f"default of {argument} from {self.extra_defaults[model][argument][0]}: "
                    )
                else:
                    # Use the same method, that was used for the CLI generation
                    argument_name = PydanticField(argument, fields[argument]).arg_names()
                    source = f"argument {', '.join(argument_name)}: "

            try:
                error_msg = str(e["ctx"]["error"])
            except KeyError:
                error_msg = e["msg"]

            msg += f"{source}{error_msg}\n"

        # Check whether parser should exit
        if self.exit_on_error:
            self.exit(ArgumentParser.EXIT_ERROR, msg)

        # Raise Error
        raise argparse.ArgumentError(None, msg)

    def error(self, message: str) -> NoReturn:
        """Prints a usage message to `stderr` and exits if required.

        Args:
            message (str): Message to print to the user.

        Raises:
            argparse.ArgumentError: Raised if not exiting on error.
            SystemExit: Raised if exiting on error.
        """
        # Print usage message
        self.print_usage(sys.stderr)

        msg = f"error: {message}\n"

        # TODO: Investigate why this function is called twice when exit_on_error is respected
        # Check whether parser should exit
        # if self.exit_on_error:
        self.exit(ArgumentParser.EXIT_ERROR, msg)

        # Raise Error
        # raise argparse.ArgumentError(None, msg)

    def _commands(self) -> argparse._SubParsersAction:  # type: ignore
        """Creates and Retrieves Subcommands Action for the ArgumentParser.

        Returns:
            argparse._SubParsersAction: SubParsersAction for the subcommands.
        """
        # Check for Existing Sub-Commands Group
        if self._subcommands is None:
            # Add Sub-Commands Group
            self._subcommands = self.add_subparsers(
                title=ArgumentParser.COMMANDS,
                action=actions.SubParsersAction,
                required=True,
            )

            # Shuffle Group to the Top for Help Message
            self._action_groups.insert(0, self._action_groups.pop())

        # Return
        return self._subcommands

    def _add_help_flag(self) -> None:
        """Adds help flag to argparser."""
        # Add help flag
        self._help_group.add_argument(
            "-h",
            "--help",
            action=argparse._HelpAction,
            help="show this help message and exit",
        )

    def _add_version_flag(self) -> None:
        """Adds version flag to argparser."""
        # Add version flag
        self._help_group.add_argument(
            "-v",
            "--version",
            action=argparse._VersionAction,
            help="show program's version number and exit",
        )

    def _add_model(
        self,
        model: type[BaseModel],
        arg_group: argparse._ArgumentGroup | None = None,
    ) -> None:
        """Adds the `pydantic` model to the argument parser.

        Args:
            model (Type[BaseModel]): Pydantic model class to add to the
                argument parser.
            arg_group: (Optional[argparse._ArgumentGroup]): argparse ArgumentGroup.
                This should not normally be passed manually, but only during
                recursion if the original model is a nested pydantic model. These
                nested models are then parsed as argument groups.
        """
        # Initialise validators dictionary
        parser = self if arg_group is None else arg_group

        explicit_groups = {}

        # Loop through fields in model
        for field in PydanticField.parse_model(model):
            if field.is_a(BaseModel):
                if field.is_subcommand():
                    command.parse_field(self._commands(), field, self.extra_defaults)
                else:
                    # for any nested pydantic models, set default factory to model_construct
                    # method. This allows pydantic to handle if no arguments from a nested
                    # submodel are passed by creating the default submodel.
                    # This is not allowed for subcommands.
                    if field.info.default_factory is None:
                        field.info.default_factory = field.model_type.model_construct

                    # create new arg group
                    group_name = field.info.title or field.name
                    arg_group = self.add_argument_group(group_name)

                    # recurse and parse fields below this submodel
                    self._add_model(model=field.model_type, arg_group=arg_group)
            else:
                # Add field
                added = False

                if (
                    self.extra_defaults is not None
                    and model in self.extra_defaults
                    and field.name in self.extra_defaults[model]
                ):
                    field.extra_default = self.extra_defaults[model][field.name]

                if isinstance(field.info, ArgFieldInfo) and field.info.hidden:
                    continue

                if isinstance(field.info, ArgFieldInfo) and field.info.group is not None:
                    if field.info.group not in explicit_groups:
                        explicit_groups[field.info.group] = self.add_argument_group(
                            field.info.group
                        )

                    parsers.add_field(explicit_groups[field.info.group], field)
                    added = True

                if not added:
                    parsers.add_field(parser, field)



# SPDX-FileCopyrightText: Hayden Richards
#
# SPDX-License-Identifier: MIT

"""Parses Standard Pydantic Fields to Command-Line Arguments.

The `standard` module contains the `parse_field` function, which parses
standard `pydantic` model fields to `ArgumentParser` command-line arguments.

Unlike the other `parser` modules, the `standard` module does not contain a
`should_parse` function. This is because it is the fallback case, where fields
that do not match any other types and require no special handling are parsed.
"""

import argparse
from typing import Any

from gallia.pydantic_argparse.utils.pydantic import PydanticField

from .utils import SupportsAddArgument


def parse_field(
    parser: SupportsAddArgument,
    field: PydanticField,
) -> None:
    """Adds standard pydantic field to argument parser.

    Args:
        parser (argparse.ArgumentParser): Argument parser to add to.
        field (PydanticField): Field to be added to parser.
    """
    args: dict[str, Any] = {}
    args.update(field.arg_required())
    args.update(field.arg_default())
    args.update(field.arg_const())
    args.update(field.arg_dest())

    # Add Standard Field
    parser.add_argument(
        *field.arg_names(),
        action=argparse._StoreAction,
        help=field.description(),
        metavar=field.metavar(),
        **args,
    )



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import reprlib
import sys
from typing import Any

from gallia.command import UDSScanner
from gallia.command.config import Field, Ranges, Ranges2D
from gallia.command.uds import UDSScannerConfig
from gallia.log import get_logger
from gallia.services.uds import NegativeResponse, UDSRequestConfig, UDSResponse
from gallia.services.uds.core.exception import IllegalResponse, UnexpectedNegativeResponse
from gallia.services.uds.core.utils import g_repr
from gallia.services.uds.helpers import suggests_sub_function_not_supported

logger = get_logger(__name__)


class ResetScannerConfig(UDSScannerConfig):
    sessions: Ranges | None = Field(
        None, description="Set list of sessions to be tested; all if None"
    )
    skip: Ranges2D = Field(
        {},
        metavar="SESSION_ID:ID",
        description="The sub functions to be skipped per session.\nA session specific skip is given by <session_id>:<sub_functions>\nwhere <sub_functions> is a comma separated list of single ids or id ranges using a dash.\nExamples:\n - 0x01:0xf3\n - 0x10-0x2f\n - 0x01:0xf3,0x10-0x2f\nMultiple session specific skips are separated by space.\nOnly takes affect if --sessions is given.\n",
    )
    check_session: int | None = Field(
        None,
        description="Check current session via read DID [for every nth subFunction] and try to recover session; only takes affect if --sessions is given",
        const=1,
    )


class ResetScanner(UDSScanner):
    """Scan ecu_reset"""

    CONFIG_TYPE = ResetScannerConfig
    SHORT_HELP = "identifier scan in ECUReset"

    def __init__(self, config: ResetScannerConfig):
        super().__init__(config)
        self.config: ResetScannerConfig = config

    async def main(self) -> None:
        if self.config.sessions is None:
            if not await self.perform_scan():
                sys.exit(1)
        else:
            sessions = self.config.sessions
            logger.info(f"testing sessions {g_repr(sessions)}")
            clean_returns = True

            # TODO: Unified shortened output necessary here
            logger.info(f"skipping identifiers {reprlib.repr(self.config.skip)}")

            for session in sessions:
                logger.notice(f"Switching to session {g_repr(session)}")
                resp: UDSResponse = await self.ecu.set_session(session)
                if isinstance(resp, NegativeResponse):
                    logger.warning(f"Switching to session {g_repr(session)} failed: {resp}")
                    continue

                logger.result(f"Scanning in session: {g_repr(session)}")
                clean_returns = clean_returns and await self.perform_scan(session)

                await self.ecu.leave_session(session, sleep=self.config.power_cycle_sleep)

            if not clean_returns:
                sys.exit(1)

    async def perform_scan(self, session: None | int = None) -> bool:
        l_ok: list[int] = []
        l_timeout: list[int] = []
        l_error: list[Any] = []

        for sub_func in range(0x01, 0x80):
            if session in self.config.skip and (
                (session_skip := self.config.skip[session]) is None or sub_func in session_skip
            ):
                logger.notice(f"skipping subFunc: {g_repr(sub_func)} because of --skip")
                continue

            if (
                session is not None
                and self.config.check_session is not None
                and (sub_func % self.config.check_session == 0)
            ):
                # Check session and try to recover from wrong session (max 3 times), else skip session
                if not await self.ecu.check_and_set_session(session):
                    logger.error(
                        f"Aborting scan on session {g_repr(session)}; current sub-func was {g_repr(sub_func)}"
                    )
                    return False

            try:
                try:
                    resp = await self.ecu.ecu_reset(
                        sub_func, config=UDSRequestConfig(tags=["ANALYZE"])
                    )
                    if isinstance(resp, NegativeResponse):
                        if suggests_sub_function_not_supported(resp):
                            logger.info(f"{g_repr(sub_func)}: {resp}")
                        else:
                            l_error.append({sub_func: resp.response_code})
                            msg = f"{g_repr(sub_func)}: with error code: {resp}"
                            logger.result(msg)
                        continue
                except IllegalResponse as e:
                    logger.warning(f"{g_repr(e)}")

                logger.result(f"{g_repr(sub_func)}: reset level found!")
                l_ok.append(sub_func)
                logger.info("Waiting for the ECU to recover…")
                await self.ecu.wait_for_ecu()

                logger.info("Reboot ECU to restore default conditions")
                resp = await self.ecu.ecu_reset(0x01)
                if isinstance(resp, NegativeResponse):
                    logger.warning(
                        f"Could not reboot ECU after testing reset level {g_repr(sub_func)}"
                    )
                else:
                    await self.ecu.wait_for_ecu()

            except TimeoutError:
                l_timeout.append(sub_func)
                if not self.config.power_cycle:
                    logger.error(f"ECU did not respond after reset level {g_repr(sub_func)}; exit")
                    sys.exit(1)

                logger.warning(
                    f"ECU did not respond after reset level {g_repr(sub_func)}; try power cycle…"
                )
                try:
                    await self.ecu.power_cycle(self.config.power_cycle_sleep)
                    await self.ecu.wait_for_ecu()
                except (TimeoutError, ConnectionError) as e:
                    logger.error(f"Failed to recover ECU: {g_repr(e)}; exit")
                    sys.exit(1)
            except ConnectionError:
                msg = f"{g_repr(sub_func)}: lost connection to ECU (post), current session: {g_repr(session)}"
                logger.warning(msg)
                await self.ecu.reconnect()
                continue

            # We reach this code only for positive responses
            # If check_session is set to any value, check here regardless of modulo
            if session is not None and self.config.check_session is not None:
                try:
                    current_session = await self.ecu.read_session()
                    logger.info(
                        f"{g_repr(sub_func)}: Currently in session {g_repr(current_session)}, "
                        f"should be {g_repr(session)}"
                    )
                except UnexpectedNegativeResponse as e:
                    logger.warning(f"Could not read current session: {e.RESPONSE_CODE.name}")

            if session is not None:
                logger.info(f"Setting session {g_repr(session)}")
                await self.ecu.set_session(session)

        logger.result(f"ok: {l_ok}")
        logger.result(f"timeout: {l_timeout}")
        logger.result(f"with error: {l_error}")

        return True



# SPDX-FileCopyrightText: Hayden Richards
#
# SPDX-License-Identifier: MIT

"""Parses Container Pydantic Fields to Command-Line Arguments.

The `container` module contains the `should_parse` function, which checks
whether this module should be used to parse the field, as well as the
`parse_field` function, which parses container `pydantic` model fields to
`ArgumentParser` command-line arguments.
"""

import argparse
import collections.abc
import enum
from typing import Any

from gallia.pydantic_argparse.utils.pydantic import PydanticField

from .utils import SupportsAddArgument


def should_parse(field: PydanticField) -> bool:
    """Checks whether the field should be parsed as a `container`.

    Args:
        field (PydanticField): Field to check.

    Returns:
        bool: Whether the field should be parsed as a `container`.
    """
    # Check and Return
    return field.is_a(collections.abc.Container) and not field.is_a((enum.Enum, str, bytes))


def parse_field(
    parser: SupportsAddArgument,
    field: PydanticField,
) -> None:
    """Adds container pydantic field to argument parser.

    Args:
        parser (argparse.ArgumentParser): Argument parser to add to.
        field (PydanticField): Field to be added to parser.
    """
    args: dict[str, Any] = {}
    args.update(field.arg_required())
    args.update(field.arg_default())
    args.update(field.arg_const())
    args.update(field.arg_dest())

    parser.add_argument(
        *field.arg_names(),
        action=argparse._StoreAction,
        nargs=argparse.ZERO_OR_MORE,
        help=field.description(),
        metavar=field.metavar(),
        **args,
    )



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import asyncio
import sys
from typing import Any

from gallia.command import UDSScanner
from gallia.command.config import AutoInt, Field, Ranges
from gallia.command.uds import UDSScannerConfig
from gallia.log import get_logger
from gallia.services.uds import NegativeResponse, UDSErrorCodes, UDSRequestConfig
from gallia.services.uds.core.constants import DiagnosticSessionControlSubFuncs, EcuResetSubFuncs
from gallia.services.uds.core.service import DiagnosticSessionControlResponse
from gallia.services.uds.core.utils import g_repr

logger = get_logger(__name__)


class SessionsScannerConfig(UDSScannerConfig):
    depth: AutoInt = Field(4, description="Specify max scanning depth.")
    sleep: AutoInt = Field(
        0,
        description="Sleep this amount of seconds after changing to DefaultSession",
        metavar="SECONDS",
    )
    skip: Ranges = Field(
        [], description="List with session IDs to skip while scanning", metavar="SESSION_ID"
    )
    with_hooks: bool = Field(False, description="Use hooks in case of a ConditionsNotCorrect error")
    reset: AutoInt | None = Field(
        None,
        description="Reset the ECU before each iteration with the optionally given reset level",
        const=0x01,
    )
    thorough: bool = Field(
        False,
        description="Perform a session scan for each 'stack' of a session instead of only once per session",
    )


class SessionsScanner(UDSScanner):
    """Iterate Sessions"""

    CONFIG_TYPE = SessionsScannerConfig
    SHORT_HELP = "session scan on an ECU"

    def __init__(self, config: SessionsScannerConfig):
        super().__init__(config)
        self.config: SessionsScannerConfig = config

        self.result: list[int] = []

    async def set_session_with_hooks_handling(
        self, session: int, use_hooks: bool
    ) -> NegativeResponse | DiagnosticSessionControlResponse:
        """This function attempts to set the session without hooks first and if unsuccessful due to NRC CnC
        it automatically attempts the session change with hooks, as long as `use_hooks` is True"""
        resp = await self.ecu.set_session(
            session, config=UDSRequestConfig(skip_hooks=True), use_db=False
        )

        if (
            isinstance(resp, NegativeResponse)
            and resp.response_code == UDSErrorCodes.conditionsNotCorrect
        ):
            logger.notice(f"Received conditionsNotCorrect for session {g_repr(session)}")
            if not use_hooks:
                logger.warning(
                    f"Session {g_repr(session)} is potentially available but could not be entered. Use --with-hooks to try to enter the session using hooks to scan for transitions available from that session."
                )
                return resp

            logger.debug(f"Attempting to change to session {g_repr(session)} with hooks")
            resp_ = await self.ecu.set_session(
                session, config=UDSRequestConfig(skip_hooks=False), use_db=False
            )

            if not isinstance(resp_, NegativeResponse):
                logger.notice(f"Successfully changed to session {g_repr(session)} with hooks")
                resp = resp_
            else:
                logger.notice(
                    f"Could not successfully change to session {g_repr(session)} even with hooks"
                )

        return resp

    async def _recover_stack(self, stack: list[int], use_hooks: bool) -> bool:
        logger.debug(f"Recovering the current session stack: {stack}")

        for session in stack:
            try:
                resp = await self.set_session_with_hooks_handling(session, use_hooks)

                if isinstance(resp, NegativeResponse):
                    logger.error(
                        f"Could not change to session {g_repr(session)} as part of stack: {resp}. Try with --reset to reset between each iteration."
                    )
                    return False

                if session == DiagnosticSessionControlSubFuncs.defaultSession:
                    logger.debug(
                        f"Sleeping for {self.config.sleep}s after changing to DefaultSession"
                    )
                    await asyncio.sleep(self.config.sleep)

            except Exception as e:
                logger.error(
                    f"Could not change to session {g_repr(session)} as part of stack: {g_repr(e)}. Try with --reset to reset between each iteration."
                )
                return False

        logger.info("Stack fully recovered!")
        return True

    async def main(self) -> None:
        found: dict[int, list[list[int]]] = {
            0: [[DiagnosticSessionControlSubFuncs.defaultSession]]
        }  # Content of this dict is just a filler to start the loop
        positive_results: list[dict[str, Any]] = []
        negative_results: list[dict[str, Any]] = []
        activated_sessions: set[int] = set()
        searched_sessions: list[int] = []

        sessions = list(range(1, 0x80))
        current_depth = 0

        while current_depth < self.config.depth and len(found[current_depth]) > 0:
            current_depth += 1

            found[current_depth] = []
            logger.notice(f"Enumerating at depth: {current_depth}")

            for stack in found[current_depth - 1]:
                if (not self.config.thorough) and stack[-1] in searched_sessions:
                    continue
                searched_sessions.append(stack[-1])

                if stack:
                    logger.info(f"Starting from session: {g_repr(stack[-1])}")

                # Recover stack in first loop iteration and afterwards only if needed
                recover_stack = True

                for session in sessions:
                    if session in self.config.skip:
                        logger.info(f"Skipping testing for session {g_repr(session)} as requested")
                        continue

                    if self.config.reset:
                        try:
                            logger.info("Resetting the ECU as requested")
                            reset_resp = await self.ecu.ecu_reset(self.config.reset)

                            if isinstance(reset_resp, NegativeResponse):
                                logger.warning(
                                    f"Could not reset ECU with {(EcuResetSubFuncs(self.config.reset).name if self.config.reset in iter(EcuResetSubFuncs) else self.config.reset)}: {reset_resp}; continuing without reset"  # type: ignore[operator]
                                )
                            else:
                                logger.info("Waiting for the ECU to recover…")
                                await self.ecu.wait_for_ecu(timeout=self.config.timeout)
                        except (TimeoutError, ConnectionError):
                            logger.warning(
                                "Lost connection to the ECU after performing a reset. Attempting to reconnect…"
                            )
                            await self.ecu.reconnect()
                        recover_stack = True

                    if recover_stack is True:
                        logger.info("Recovering the stack...")
                        if not await self._recover_stack(stack, self.config.with_hooks):
                            sys.exit(1)
                        recover_stack = False

                    logger.debug(f"Attempting to change to session {session:#04x}")
                    try:
                        resp = await self.set_session_with_hooks_handling(
                            session, self.config.with_hooks
                        )

                        # Do only ignore NCR subFunctionNotSupported
                        if (
                            isinstance(resp, NegativeResponse)
                            and resp.response_code == UDSErrorCodes.subFunctionNotSupported
                        ):
                            logger.info(f"Could not change to session {g_repr(session)}: {resp}")
                            # For this NRC, continue without recovering the stack
                            continue

                        logger.notice(
                            f"Found session: {g_repr(session)} via stack: {g_repr(stack)}; {resp}"
                        )

                        if isinstance(resp, NegativeResponse):
                            negative_results.append(
                                {"session": session, "stack": stack, "error": resp.response_code}
                            )
                            # Presumably we did not successfully leave the session, so no need to recover the stack
                            continue

                        # Do not track a session in "found" if it is already present on the stack unless 'thorough'
                        # This avoids looping through sessions, e.g. 0x01->0x02->0x01->0x02->...
                        if self.config.thorough or session not in stack:
                            found[current_depth].append(stack + [session])

                        activated_sessions.add(session)
                        positive_results.append({"session": session, "stack": stack, "error": None})
                    except TimeoutError:
                        logger.warning(f"Could not change to session {g_repr(session)}: Timeout")
                        # TODO: Is there a need to recover stack in this case?
                        continue
                    except Exception as e:
                        logger.warning(
                            f"Mamma mia: {repr(e)} for session {session:#04x} from stack {stack}"
                        )

                    # If the loop is not `continue`d early, recover stack, e.g. on successful session change
                    recover_stack = True

        if len(positive_results) > 0:
            logger.result("Scan finished; Found the following sessions:")
            previous_session = 0

            for res in sorted(positive_results, key=lambda x: x["session"]):
                session = res["session"]

                if session != previous_session:
                    previous_session = session
                    self.result.append(int(session))
                    logger.result(f"* Session {g_repr(session)} ")

                    if self.db_handler is not None:
                        await self.db_handler.insert_session_transition(session, res["stack"])

                logger.result(f"\tvia stack: {'->'.join([f'{g_repr(i)}' for i in res['stack']])}")
        else:
            logger.result("Could not successfully change to any session!")

        if len(negative_results) > 0:
            logger.result("The following sessions were identified but could not be activated:")
            previous_session = 0

            for res in sorted(negative_results, key=lambda x: x["session"]):
                session = res["session"]

                if (
                    session not in activated_sessions
                    and res["error"] != UDSErrorCodes.subFunctionNotSupportedInActiveSession
                ):
                    if session != previous_session:
                        previous_session = session
                        logger.result(f"* Session {g_repr(session)} ")

                        if self.db_handler is not None:
                            await self.db_handler.insert_session_transition(session, res["stack"])

                    logger.result(
                        f"\tvia stack: {'->'.join([f'{g_repr(i)}' for i in res['stack']])} (NRC: {res['error']})"
                    )
